''' Load visual data from text files.'''
from bluesky.navdatabase.loadnavdata_txt import thresholds, thrpoints
from zipfile import ZipFile
import numpy as np
import bluesky as bs


REARTH_INV = 1.56961231e-7

# used for calculating the vertices of the runways as well as the threshold boxes
def dlatlon(lat0, lon0, lat1, lon1, width):

    # calculate distance between ends of runways / threshold boxes
    flat_earth = np.cos(0.5 * np.radians(lat0 + lat1))
    lx = lat1 - lat0
    ly = (lon1 - lon0) * flat_earth
    l = np.sqrt(lx * lx + ly * ly)
    wx = ly / l * 0.5 * width
    wy = -lx / l * 0.5 * width
    dlat = np.degrees(wx * REARTH_INV)
    dlon = np.degrees(wy * REARTH_INV / flat_earth)

    # store the vertice information per runway /threshold box
    vertices = [lat0 + dlat, lon0 + dlon,
                lat0 - dlat, lon0 - dlon,
                lat1 + dlat, lon1 + dlon,
                lat0 - dlat, lon0 - dlon,
                lat1 + dlat, lon1 + dlon,
                lat1 - dlat, lon1 - dlon]

    return vertices

def threshold_vertices(lat, lon, bearing):
    ''' Generate threshold vectices for given runway threshold data. '''
    d_box = 20.0 * REARTH_INV  # m
    width_box = 30  # m

    # bearing in opposite direction
    opp_bearing = np.radians((np.degrees(bearing) + 180) % 360)

    # get points at both ends of the boxes around the threshold point
    # (zebra crossing)
    latbox0, lonbox0 = thrpoints(lat, lon, d_box, bearing)
    latbox1, lonbox1 = thrpoints(lat, lon, d_box, opp_bearing)

    # calculate vertices of threshold box
    vertices = dlatlon(np.degrees(latbox0), np.degrees(lonbox0), np.degrees(latbox1),
                        np.degrees(lonbox1), width_box)

    return vertices


def load_coastline_txt():
    # -------------------------COASTLINE DATA----------------------------------
    # Init geo (coastline)  data and convert pen up/pen down format of
    # coastlines to numpy arrays with lat/lon
    coast = []
    clat = clon = 0.0
    with open(bs.resource(bs.settings.navdata_path) / 'coastlines.dat', 'r') as f:
        print("Reading coastlines.dat")
        for line in f:
            line = line.strip()
            if not (line == "" or line[0] == '#'):
                arg = line.split()
                if len(arg) == 3:
                    lat, lon = float(arg[1]), float(arg[2])
                    if arg[0] == 'D':
                        coast.append([clat, clon, lat, lon])
                    clat, clon = lat, lon
    # Sort the line segments by longitude of the first vertex
    coastvertices = np.array(
        sorted(coast, key=lambda a_entry: a_entry[1]), dtype=np.float32)
    coastindices = np.zeros(361)
    coastlon = coastvertices[:, 1]
    for i in range(0, 360):
        coastindices[i] = np.searchsorted(coastlon, i - 180) * 2
    coastvertices.resize((int(coastvertices.size / 2), 2))
    del coast
    return coastvertices, coastindices


# Only try this if BlueSky is started in qtgl gui mode
if bs.gui == 'qtgl':
    from PyQt6.QtCore import Qt
    from PyQt6.QtWidgets import QApplication, QProgressDialog
    from bluesky.ui.polytools import PolygonSet, BoundingBox


    class ProgressBar():
        def __init__(self, text):
            if QApplication.instance() is None:
                self.dialog = None
                print(text)
            else:
                self.dialog = QProgressDialog(text, 'Cancel', 0, 100)
                # self.dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
                self.dialog.setModal(True)
                self.dialog.show()

        def update(self, value):
            if self.dialog:
                self.dialog.setValue(int(value))
                QApplication.processEvents()
            else:
                print('Progress: %.2f%% done' % value)

        def close(self):
            if self.dialog:
                self.dialog.close()


    """ Load static data for GUI from files such as airport, shapes, etc."""

    def load_aptsurface_txt():
        """ Read airport data from navigation database files"""
        pb = ProgressBar('Binary buffer file not found, or file out of date: Constructing vertex buffers from geo data.')

        runways = []
        rwythr = []
        asphalt = PolygonSet()
        concrete = PolygonSet()
        cur_poly = asphalt
        apt_indices = []
        apt_ctr_lat = []
        apt_ctr_lon = []
        apt_bb = BoundingBox()
        count = 0
        bytecount = 0
        zfile = ZipFile(bs.resource(bs.settings.navdata_path) / 'apt.zip')
        fsize = float(zfile.getinfo('apt.dat').file_size)
        print("Reading apt.dat from apt.zip")
        with zfile.open('apt.dat', 'r') as f:
            for line in f:
                bytecount += len(line)
                # Check how far we are
                count += 1
                if count % 1000 == 0:
                    pb.update((bytecount / fsize * 100.0))

                elems = line.decode(encoding="ascii", errors="ignore").strip().split()
                if not elems:
                    continue

                # 1: AIRPORT
                if elems[0] in ['1', '16', '17']: # 1=airport, 16=seaplane base, 17=heliport
                    cur_poly.end()
                    # Store the starting vertices for this apt, [0, 0] if this is the first apt
                    if apt_indices:
                        start_indices = [apt_indices[-1][0] + apt_indices[-1][1],
                                         apt_indices[-1][2] + apt_indices[-1][3]]
                    else:
                        start_indices = [0, 0]

                    if asphalt.bufsize() // 2 > start_indices[0] or concrete.bufsize() // 2 > start_indices[1]:
                        apt_indices.append( [
                                                start_indices[0],
                                                asphalt.bufsize() // 2 - start_indices[0],
                                                start_indices[1],
                                                concrete.bufsize() // 2 - start_indices[1]
                                            ])

                        center = apt_bb.center()
                        apt_ctr_lat.append(center[0])
                        apt_ctr_lon.append(center[1])

                    # Reset the boundingbox
                    apt_bb = BoundingBox()
                    continue

                # 100: LAND RUNWAY
                if elems[0] == '100':
                    width = float(elems[1])
                    # Only asphalt and concrete runways
                    if int(elems[2]) > 2:
                        continue
                    # rwy_lbl = (elems[8], elems[17])

                    lat0 = float(elems[9])
                    lon0 = float(elems[10])
                    offset0 = float(elems[11])

                    lat1 = float(elems[18])
                    lon1 = float(elems[19])
                    offset1 = float(elems[20])

                    # runway vertices
                    runways.extend(dlatlon(lat0, lon0, lat1, lon1, width))

                    # threshold information: ICAO code airport, Runway identifier,
                    # latitude, longitude, bearing
                    # vertices: gives vertices of the box around the threshold

                    # opposite runways are on the same line. RWY1: 8-11, RWY2: 17-20
                    # Hence, there are two thresholds per line
                    # thr0: First lat0 and lon0 , then lat1 and lat1, offset=[11]
                    # thr1: First lat1 and lat1 , then lat0 and lon0, offset=[20]
                    thr0 = thresholds(np.radians(lat0), np.radians(lon0), np.radians(lat1), np.radians(lon1), offset0)
                    vertices0 = threshold_vertices(*thr0)
                    thr1 = thresholds(np.radians(lat1), np.radians(lon1), np.radians(lat0), np.radians(lon0), offset1)
                    vertices1 = threshold_vertices(*thr1)
                    rwythr.extend(vertices0)
                    rwythr.extend(vertices1)

                    continue

                # 110: TAXIWAY/PAVEMENT: Start of polygon contour
                if elems[0] == '110':
                    cur_poly.end()
                    if elems[1] == '1':
                        cur_poly = asphalt
                        cur_poly.begin()
                    elif elems[1] == '2':
                        cur_poly = concrete
                        cur_poly.begin()
                    continue

                # 130: AIRPORT BOUNDARY
                if elems[0] == '130':
                    cur_poly.end()
                    continue

                controlpoint = None
                # Straight line (111) or bezier line (112) in contour
                if elems[0] == '111' or elems[0] == '112':
                    vertex = np.array((float(elems[1]), float(elems[2]), 0.0))
                    apt_bb.update(vertex)

                    if cur_poly.in_poly:
                        if elems[0] == '112':
                            controlpoint = np.array((float(elems[3]), float(elems[4]), 0.0))

                        cur_poly.addVertex(vertex, controlpoint)

                # Straight (113) or bezier (114) contour endpoint
                elif elems[0] == '113' or elems[0] == '114':
                    vertex = np.array((float(elems[1]), float(elems[2]), 0.0))
                    apt_bb.update(vertex)

                    if cur_poly.in_poly:
                        if elems[0] == '114':
                            controlpoint = np.array((float(elems[3]), float(elems[4]), 0.0))

                        cur_poly.addVertex(vertex, controlpoint)

                        cur_poly.endContour()

                else:
                    cur_poly.end()

        # calculate the location of the runway thresholds

        # Clean up:
        cur_poly.end()
        start_indices = [apt_indices[-1][0] + apt_indices[-1][1],
                         apt_indices[-1][2] + apt_indices[-1][3]]

        if asphalt.bufsize() > start_indices[0] or concrete.bufsize() > start_indices[1]:
            apt_indices.append( [
                                    start_indices[0],
                                    int(asphalt.bufsize() / 2) - start_indices[0],
                                    start_indices[1],
                                    int(concrete.bufsize() / 2) - start_indices[1]
                                ])

            center = apt_bb.center()
            apt_ctr_lat.append(center[0])
            apt_ctr_lon.append(center[1])

        # Store in binary pickle file
        vbuf_asphalt = np.array(asphalt.vbuf, dtype=np.float32)
        vbuf_concrete = np.array(concrete.vbuf, dtype=np.float32)
        vbuf_runways = np.array(runways, dtype=np.float32)
        vbuf_rwythr = np.array(rwythr, dtype=np.float32)
        apt_ctr_lat = np.array(apt_ctr_lat)
        apt_ctr_lon = np.array(apt_ctr_lon)
        apt_indices = np.array(apt_indices)

        # Close the progress dialog
        pb.close()

        # return the data
        return vbuf_asphalt, vbuf_concrete, vbuf_runways, vbuf_rwythr, apt_ctr_lat, apt_ctr_lon, apt_indices



import inspect
from collections import defaultdict
from typing import Callable, Optional

import bluesky as bs
from bluesky.core.funcobject import FuncObject
from bluesky.core.timedfunction import timed_function
from bluesky.core.walltime import Timer
from bluesky.network.common import ActionType
from bluesky.network.subscriber import subscriber
from bluesky.network.sharedstate import _recursive_update
import bluesky.network.context as ctx


class PublisherMeta(type):
    __publishers__ = dict()
    __timers__ = dict()

    def __call__(cls, topic: str, dt=None, collect=False, send_type='replace'):
        pub = PublisherMeta.__publishers__.get(topic)
        if pub is None:
            pub = PublisherMeta.__publishers__[topic] = super().__call__(topic, dt, collect)
            # If dt is specified, also create a timer
            if dt is not None:
                if dt not in PublisherMeta.__timers__:
                    timer = PublisherMeta.__timers__[dt] = Timer(dt)
                else:
                    timer = PublisherMeta.__timers__[dt]
                callback = getattr(pub, f'send_{send_type}')
                timer.timeout.connect(callback)
        return pub

    @subscriber
    @staticmethod
    def request(*topics):
        ''' Clients can request a full state update, to which this function responds. '''
        for pub in (p for p in map(PublisherMeta.__publishers__.get, topics) if p is not None):
            pub.send_replace(to_group=ctx.sender_id)


class StatePublisher(metaclass=PublisherMeta):
    ''' BlueSky shared state publisher class.

        Use this class if you want more control over publishing shared state data.
        Unlike the state_publisher decorator this class allows you to also send 
        update, delete, append, and extend actions.
    '''
    __collect__ = defaultdict(list)

    @classmethod
    def collect(cls, topic: str, payload: list, to_group: str=''):
        if payload[0] == ActionType.Replace:
            cls.__collect__[(topic, to_group)] = payload
            return
        store = cls.__collect__[(topic, to_group)]
        if not store or store[-2] not in (payload[0], ActionType.Replace.value, ActionType.Extend.value):
            if payload[0] == ActionType.Append.value:
                payload[0] = ActionType.Extend.value
                payload[1] = {k:[v] for k, v in payload[1].items()}
            store.extend(payload)
        elif payload[0] == ActionType.Update.value:
            _recursive_update(store[1], payload[1])
        elif payload[0] == ActionType.Append.value:
            for key, item in payload[1].items():
                store[1][key].append(item)
        elif payload[0] == ActionType.Extend.value:
            for key, item in payload[1].items():
                store[1][key].extend(item)

    @staticmethod
    @timed_function(hook=('update', 'hold'))
    def send_collected():
        while StatePublisher.__collect__:
            (topic, to_group), payload = StatePublisher.__collect__.popitem()
            bs.net.send(topic, payload, to_group)

    def __init__(self, topic: str, dt=None, collect=False, send_type='replace') -> None:
        self.topic = topic
        self.dt = dt
        self.collects = collect

    @staticmethod
    def get_payload():
        ''' Default payload function returns None '''
        return

    def send_update(self, to_group=b'', **data):
        data = data or self.get_payload()
        if data:
            if self.collects:
                self.collect(self.topic, [ActionType.Update.value, data], to_group)
            else:
                bs.net.send(self.topic, [ActionType.Update.value, data], to_group)


    def send_delete(self, to_group=b'', **keys):
        if keys:
            bs.net.send(self.topic, [ActionType.Delete.value, keys], to_group)


    def send_append(self, to_group=b'', **data):
        data = data or self.get_payload()
        if data:
            if self.collects:
                self.collect(self.topic, [ActionType.Append.value, data], to_group)
            else:
                bs.net.send(self.topic, [ActionType.Append.value, data], to_group)

    def send_extend(self, to_group=b'', **data):
        data = data or self.get_payload()
        if data:
            if self.collects:
                self.collect(self.topic, [ActionType.Extend.value, data], to_group)
            else:
                bs.net.send(self.topic, [ActionType.Extend.value, data], to_group)

    def send_replace(self, to_group=b'', **data):
        data = data or self.get_payload()
        if data:
            bs.net.send(self.topic, [ActionType.Replace.value, data], to_group)

    def payload(self, func: Callable):
        ''' Decorator method to specify payload getter for this Publisher. '''
        self.get_payload = FuncObject(func)
        return func


def state_publisher(func: Optional[Callable] = None, *, topic='', dt=None, send_type='replace'):
    ''' BlueSky shared state publisher decorator.

        Convenience decorator to create a periodically called state publisher.
        Functions decorated with this decorator will be:
        - periodically called at interval dt
        - called when a subscriber requests a full state

        Decorated function should return a dictionary with all relevant state data.

        Arguments:
        - func: The payload function that returns the data to be published in a dictionary
        - topic: The message topic that clients can subscribe to. If no topic is given, the name of the
          decorated function is used.
        - dt: Time interval between successive periodic broadcasts of this state publisher
        - send_type: Type of update sent by this publisher. This can be:
          - replace (the default): Each state package sent by this publisher fully replaces previous data
          - append: Each state package sent should be appended to the previously received data
          - extend: Each state package sent should be used to extend the previously received data
          - update: Each state package sent should be used to update the previously received data
    '''
    def deco(func):
        ifunc = inspect.unwrap(func, stop=lambda f:not isinstance(func, (staticmethod, classmethod)))
        itopic = (topic or ifunc.__name__).upper()

        StatePublisher(itopic, dt, send_type=send_type).payload(func)
        return func

    # Allow both @publisher and @publisher(args)
    return deco if func is None else deco(func)


from math import cos, atan2, radians, degrees
from numpy import array
import bluesky as bs

from bluesky.tools import geo
from bluesky.tools.misc import findnearest, cmdsplit


def radarclick(cmdline, lat, lon, acdata=None, route=None):
    """Process lat,lon as clicked in radar window"""

    # Specify which argument can be clicked, and how, in this dictionary
    # and when it's the last, also add ENTER

    clickcmd = {"": "acid,-",
                "ADDWPT": "acid,latlon,-,-,wpinroute,-",
                "AFTER": "acid,wpinroute,-",
                "AT": "acid,wpinroute,-",
                "ALT": "acid,-",
                "AREA": "latlon,-,latlon",
                "ASAS": "acid,-",
                "BOX": "-,latlon,-,latlon",
                "CIRCLE": "-,latlon,-,dist",
                "CRE":  "-,-,latlon,-,hdg,-,-",
                "DEFWPT": "-,latlon,-",
                "DEL": "acid,...",
                "DELWPT": "acid,wpinroute,-",
                "DELRTE": "acid,-",
                "DEST": "acid,apt",
                "DIRECT": "acid,wpinroute",
                "DIST": "latlon,-,latlon",
                "DUMPRTE": "acid",
                "ENG": "acid,-",
                "GETWIND":"latlon,-",
                "GROUP":"-,acid,...",
                "HDG": "acid,hdg",
                "LINE": "-,latlon,-,latlon",
                "LISTRTE": "acid,-",
                "LNAV": "acid,-",
                "MOVE": "acid,latlon,-,-,hdg",
                "NAVDISP": "acid",
                "NOM": "acid",
                "ND": "acid",
                "ORIG": "acid,apt",
                "PAN": "latlon",
                "POLY": "-,latlon,...",
                "POLYALT": "-,-,-,latlon,...",
                "POLYGON": "-,latlon,...",
                "POLYLINE": "-,latlon,...",
                "POS": "acid",
                "SSD": "acid,...",
                "SPD": "acid,-",
                "TRAIL":"acid,-",
                "VNAV": "acid,-",
                "VS": "acid,-",
                "WIND":"latlon,-",
                "WINDGFS":"latlon,-,latlon,-"
                }

    # Default values, when nothing is found to be added based on click
    todisplay = ""  # Result of click is added here
    tostack   = ""  # If it is the last argument we will pass whole line to the stack

    # The pygame version has access to the complete traffic object. This gets
    # passed to radarclick in the QtGL version.
    if acdata is None:
        acdata = bs.traf

    # Split command line into command and arguments, pass traf ids to check for
    # switched acid and command
    cmd, args = cmdsplit(cmdline, acdata.id)
    cmd = cmd.upper()
    numargs   = len(args)

    # -------- Process click --------
    # Double click on aircraft = POS command
    if numargs == 0 and acdata.id.count(cmd.upper()) > 0:
        todisplay = "\n"          # Clear the current command
        tostack   = "POS " + cmd  # And send a pos command to the stack

    # Insert: nearest aircraft id
    else:

        # TODO: Check for synonyms (dictionary is imported from stack)
        # if cmd in cmdsynon:
        #    cmd = cmdsynon[cmd.upper()]

        # Try to find command in clickcmd dictionary
        try:
            lookup = clickcmd[cmd.upper()]

        except KeyError:
            # When command was not found in dictionary:
            # do nothing, return empty strings
            return "", ""

        # For valid value, insert relevant dat on edit line
        if lookup:
            if len(cmdline) > 0 and (cmdline[-1] != " " and cmdline[-1]!=","):
                todisplay = " "

            # Determine argument click type
            clickargs = lookup.lower().split(",")
            totargs   = len(clickargs)
            curarg    = numargs
            # Exception case: if the last item of the clickargs list is "..."
            # then the one-but-last can be repeatedly added
            # (e.g. for the definition of a polygon)

            if clickargs[-1] == "...":
                totargs = 999
                curarg  = min(curarg, len(clickargs) - 2)

            if curarg < totargs:
                clicktype = clickargs[curarg]

                if clicktype == "acid":
                    idx = findnearest(lat, lon, acdata.lat, acdata.lon)
                    if idx >= 0:
                        todisplay += acdata.id[idx] + " "

                elif clicktype == "latlon":
                    todisplay += str(round(lat, 6)) + "," + str(round(lon, 6)) + " "

                elif clicktype == "dist":
                    latref, lonref = float(args[1]), float(args[2])
                    todisplay += str(round(geo.kwikdist(latref, lonref, lat, lon), 6))

                elif clicktype == "apt":
                    idx = findnearest(lat, lon, bs.navdb.aptlat, bs.navdb.aptlon)
                    if idx >= 0:
                        todisplay += bs.navdb.aptid[idx] + " "

                elif clicktype == "wpinroute":  # Find nearest waypoint in route
                    if acdata.id.count(args[0]) > 0:
                        itraf      = acdata.id.index(args[0])
                        synerr = False
                        reflat = acdata.lat[itraf]
                        reflon = acdata.lon[itraf]
                        # The pygame version can get the route directly from traf
                        # otherwise the route is passed to this function
                        if route is None:
                            route = acdata.ap.route[itraf]

                        if len(route.wplat) > 0:
                            iwp = findnearest(lat, lon,
                                        array(route.wplat),
                                        array(route.wplon))
                            if iwp >= 0:
                                todisplay += route.wpname[iwp]+" "

                    else:
                        synerr = True

                elif clicktype == "hdg":
                    # Read start position from command line
                    if cmd == "CRE":
                        try:
                            reflat = float(args[2])
                            reflon = float(args[3])
                            synerr = False
                        except:
                            synerr = True
                    elif cmd == "MOVE":
                        try:
                            reflat = float(args[1])
                            reflon = float(args[2])
                            synerr = False
                        except:
                            synerr = True
                    else:
                        if acdata.id.count(args[0]) > 0:
                            idx    = acdata.id.index(args[0])
                            reflat = acdata.lat[idx]
                            reflon = acdata.lon[idx]
                            synerr = False
                        else:
                            synerr = True
                    if not synerr:
                        dy = lat - reflat
                        dx = (lon - reflon) * cos(radians(reflat))
                        hdg = degrees(atan2(dx, dy)) % 360.

                        todisplay += str(int(hdg)) + " "

                # Is it the last argument? (then we will insert ENTER as well)
                if curarg + 1 >= totargs:
                    tostack = cmdline + todisplay
                    # todisplay = todisplay + '\n'
                    todisplay = ''

    return tostack, todisplay



from bluesky import settings
from bluesky.traffic.performance.perfbase import PerfBase
try:
    import bluesky.traffic.performance.openap
    print('Successfully loaded OpenAP performance model')
except ImportError:
    print('Failed to load OpenAP performance model')
try:
    import bluesky.traffic.performance.bada
    print('Successfully loaded BADA performance model')
except ImportError:
    print('Failed to load BADA performance model')
try:
    import bluesky.traffic.performance.legacy
    print('Successfully loaded legacy performance model')
except ImportError:
    print('Failed to load BlueSky legacy performance model')

settings.set_variable_defaults(performance_model='openap')

# Set default performance model
PerfBase.setdefault(settings.performance_model)



''' This module provides the Conflict Detection base class. '''
import numpy as np

import bluesky as bs
from bluesky.tools.aero import ft, nm
from bluesky.core import Entity
from bluesky.stack import command


bs.settings.set_variable_defaults(asas_pzr=5.0, asas_pzh=1000.0,
                                  asas_dtlookahead=300.0)


class ConflictDetection(Entity, replaceable=True):
    ''' Base class for Conflict Detection implementations. '''
    def __init__(self):
        super().__init__()
        ## Default values
        # [m] Horizontal separation minimum for detection
        self.rpz_def = bs.settings.asas_pzr * nm
        self.global_rpz = True
        # [m] Vertical separation minimum for detection
        self.hpz_def = bs.settings.asas_pzh * ft
        self.global_hpz = True
        # [s] lookahead time
        self.dtlookahead_def = bs.settings.asas_dtlookahead
        self.global_dtlook = True
        self.dtnolook_def = 0.0
        self.global_dtnolook = True

        # Conflicts and LoS detected in the current timestep (used for resolving)
        self.confpairs = list()
        self.lospairs = list()
        self.qdr = np.array([])
        self.dist = np.array([])
        self.dcpa = np.array([])
        self.tcpa = np.array([])
        self.tLOS = np.array([])
        # Unique conflicts and LoS in the current timestep (a, b) = (b, a)
        self.confpairs_unique = set()
        self.lospairs_unique = set()

        # All conflicts and LoS since simt=0
        self.confpairs_all = list()
        self.lospairs_all = list()

        # Per-aircraft conflict data
        with self.settrafarrays():
            self.inconf = np.array([], dtype=bool)  # In-conflict flag
            self.tcpamax = np.array([]) # Maximum time to CPA for aircraft in conflict
            # [m] Horizontal separation minimum for detection
            self.rpz = np.array([])
            # [m] Vertical separation minimum for detection
            self.hpz = np.array([])
            # [s] lookahead time
            self.dtlookahead = np.array([])
            self.dtnolook = np.array([])

    def clearconfdb(self):
        ''' Clear conflict database. '''
        self.confpairs_unique.clear()
        self.lospairs_unique.clear()
        self.confpairs.clear()
        self.lospairs.clear()
        self.qdr = np.array([])
        self.dist = np.array([])
        self.dcpa = np.array([])
        self.tcpa = np.array([])
        self.tLOS = np.array([])
        self.inconf = np.zeros(bs.traf.ntraf)
        self.tcpamax = np.zeros(bs.traf.ntraf)

    def create(self, n):
        super().create(n)
        # Initialise values of own states
        self.rpz[-n:] = self.rpz_def
        self.hpz[-n:] = self.hpz_def
        self.dtlookahead[-n:] = self.dtlookahead_def
        self.dtnolook[-n:] = self.dtnolook_def

    def reset(self):
        super().reset()
        self.clearconfdb()
        self.confpairs_all.clear()
        self.lospairs_all.clear()
        self.rpz_def = bs.settings.asas_pzr * nm
        self.hpz_def = bs.settings.asas_pzh * ft
        self.dtlookahead_def = bs.settings.asas_dtlookahead
        self.dtnolook_def = 0.0
        self.global_rpz = self.global_hpz = True
        self.global_dtlook = self.global_dtnolook = True

    @staticmethod
    @command(name='CDMETHOD', aliases=('ASAS',))
    def setmethod(name : 'txt' = ''):
        ''' Select a Conflict Detection (CD) method. '''
        # Get a dict of all registered CD methods
        methods = ConflictDetection.derived()
        names = ['OFF' if n == 'CONFLICTDETECTION' else n for n in methods]
        if not name:
            curname = 'OFF' if ConflictDetection.selected() is ConflictDetection \
                else ConflictDetection.selected().__name__
            return True, f'Current CD method: {curname}' + \
                         f'\nAvailable CD methods: {", ".join(names)}'
        # Check if the requested method exists
        if name == 'OFF':
            # Select the base method and clear the conflict database
            ConflictDetection.select()
            ConflictDetection.instance().clearconfdb()
            return True, 'Conflict Detection turned off.'
        if name == 'ON':
            # Just select the first CD method in the list
            name = next(n for n in names if n != 'OFF')
        method = methods.get(name, None)
        if method is None:
            return False, f'{name} doesn\'t exist.\n' + \
                          f'Available CD methods: {", ".join(names)}'

        # Select the requested method
        method.select()
        ConflictDetection.instance().clearconfdb()
        return True, f'Selected {method.__name__} as CD method.'

    @command(name='ZONER', aliases=('PZR', 'RPZ', 'PZRADIUS'))
    def setrpz(self, radius: float = -1.0, *acidx: 'acid'):
        ''' Set the horizontal separation distance (i.e., the radius of the
            protected zone) in nautical miles. 
            
            Arguments:
            - radius: The protected zone radius in nautical miles
            - acidx: Aircraft id(s) or group. When this argument is not provided the default PZ radius is changed.
              Otherwise the PZ radius for the passed aircraft is changed. '''
        if radius < 0.0:
            return True, f'ZONER [radius(nm), acid(s)/ac group]\nCurrent default PZ radius: {self.rpz_def / nm:.2f} NM'
        if len(acidx) > 0:
            if isinstance(acidx[0], np.ndarray):
                acidx = acidx[0]
            self.rpz[acidx] = radius * nm
            self.global_rpz = False
            return True, f'Setting PZ radius to {radius} NM for {len(acidx)} aircraft'
        oldradius = self.rpz_def
        self.rpz_def = radius * nm
        if self.global_rpz:
            self.rpz[:] = self.rpz_def
        # Adjust factors for reso zone if those were set with an absolute value
        if not bs.traf.cr.resorrelative:
            bs.stack.stack(f"RSZONER {bs.traf.cr.resofach*oldradius/nm}")
        return True, f'Setting default PZ radius to {radius} NM'

    @command(name='ZONEDH', aliases=('PZDH', 'DHPZ', 'PZHEIGHT'))
    def sethpz(self, height: float = -1.0, *acidx: 'acid'):
        ''' Set the vertical separation distance (i.e., half of the protected
            zone height) in feet.

            Arguments:
            - height: The vertical separation height in feet
            - acidx: Aircraft id(s) or group. When this argument is not provided the default PZ height is changed.
              Otherwise the PZ height for the passed aircraft is changed. '''
        if height < 0.0:
            return True, f'ZONEDH [height (ft), acid(s)/ac group]\nCurrent default PZ height: {self.hpz / ft:.2f} ft'
        if len(acidx) > 0:
            if isinstance(acidx[0], np.ndarray):
                acidx = acidx[0]
            self.hpz[acidx] = height * ft
            self.global_hpz = False
            return True, f'Setting PZ height to {height} ft for {len(acidx)} aircraft'
        oldhpz = self.hpz_def
        self.hpz_def = height * ft
        if self.global_hpz:
            self.hpz[:] = self.hpz_def
        # Adjust factors for reso zone if those were set with an absolute value
        if not bs.traf.cr.resodhrelative:
            bs.stack.stack(f"RSZONEDH {bs.traf.cr.resofacv*oldhpz/ft}")
        return True, f'Setting default PZ height to {height} ft'

    @command(name='DTLOOK')
    def setdtlook(self, time: 'time' = -1.0, *acidx: 'acid'):
        ''' Set the lookahead time (in [hh:mm:]sec) for conflict detection. '''
        if time < 0.0:
            return True, f'DTLOOK[time]\nCurrent value: {self.dtlookahead_def: .1f} sec'
        if len(acidx) > 0:
            if isinstance(acidx[0], np.ndarray):
                acidx = acidx[0]
            self.dtlookahead[acidx] = time
            self.global_dtlook = False
            return True, f'Setting CD lookahead to {time} sec for {len(acidx)} aircraft'
        self.dtlookahead_def = time
        if self.global_dtlook:
            self.dtlookahead[:] = time
        return True, f'Setting default CD lookahead to {time} sec'

    @command(name='DTNOLOOK')
    def setdtnolook(self, time: 'time' = -1.0, *acidx: 'acid'):
        ''' Set the interval (in [hh:mm:]sec) in which conflict detection
            is skipped after a conflict resolution. '''
        if time < 0.0:
            return True, f'DTNOLOOK[time]\nCurrent value: {self.dtnolook_def: .1f} sec'
        if len(acidx) > 0:
            if isinstance(acidx[0], np.ndarray):
                acidx = acidx[0]
            self.dtnolook[acidx] = time
            self.global_dtnolook = False
            return True, f'Setting CD no-look to {time} sec for {len(acidx)} aircraft'
        self.dtnolook_def = time
        if self.global_dtnolook:
            self.dtnolook[:] = time
        return True, f'Setting default CD no-look to {time} sec'

    def update(self, ownship, intruder):
        ''' Perform an update step of the Conflict Detection implementation. '''
        self.confpairs, self.lospairs, self.inconf, self.tcpamax, self.qdr, \
            self.dist, self.dcpa, self.tcpa, self.tLOS = \
                self.detect(ownship, intruder, self.rpz, self.hpz, self.dtlookahead)

        # confpairs has conflicts observed from both sides (a, b) and (b, a)
        # confpairs_unique keeps only one of these
        confpairs_unique = {frozenset(pair) for pair in self.confpairs}
        lospairs_unique = {frozenset(pair) for pair in self.lospairs}

        self.confpairs_all.extend(confpairs_unique - self.confpairs_unique)
        self.lospairs_all.extend(lospairs_unique - self.lospairs_unique)

        # Update confpairs_unique and lospairs_unique
        self.confpairs_unique = confpairs_unique
        self.lospairs_unique = lospairs_unique

    def detect(self, ownship, intruder, rpz, hpz, dtlookahead):
        ''' Detect any conflicts between ownship and intruder.
            This function should be reimplemented in a subclass for actual
            detection of conflicts. See for instance
            bluesky.traffic.asas.statebased.
        '''
        confpairs = []
        lospairs = []
        inconf = np.zeros(ownship.ntraf)
        tcpamax = np.zeros(ownship.ntraf)
        qdr = np.array([])
        dist = np.array([])
        dcpa = np.array([])
        tcpa = np.array([])
        tLOS = np.array([])
        return confpairs, lospairs, inconf, tcpamax, qdr, dist, dcpa, tcpa, tLOS



# A Simple UDP class
import socket
import msgpack
import zmq
from bluesky.core.signal import Signal
from bluesky.network.common import get_ownip, IDLEN
from bluesky import settings

settings.set_variable_defaults(discovery_port=11000)

IS_SERVER = 0
IS_CLIENT = 1
IS_REQUEST = 2
IS_REPLY = 4

class Discovery:
    """simple UDP ping class"""
    def __init__(self, own_id, is_client=True):
        self.address = get_ownip()
        self.broadcast = '255.255.255.255'
        self.port = settings.discovery_port
        self.own_id = own_id
        self.mask = IS_CLIENT if is_client else IS_SERVER

        # When run as a client, typically our own poller is used
        # In this case, also a signal is emitted when one or more
        # servers are discovered
        self.poller = None
        self.server_discovered = Signal('server-discovered')

        # Create UDP socket
        self.handle = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

        # Ask operating system to let us do broadcasts from socket
        self.handle.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        if hasattr(socket, 'SO_REUSEPORT'):
            self.handle.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        else:
            self.handle.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Bind UDP socket to local port so we can receive pings
        self.handle.bind(('', self.port))

    def start(self):
        ''' Start UDP-based discovery of available BlueSky servers. '''
        if not self.poller:
            self.poller = zmq.Poller()
            self.poller.register(self.handle, zmq.POLLIN)
            self.send_request()

    def stop(self):
        ''' Stop UDP-based discovery. '''
        if self.poller:
            self.poller.unregister(self.handle)
            self.poller = None

    def update(self):
        ''' Periodically call this function to find remote servers. 
            The 'server-discovered' Signal is emitted whenever a
            server is found.
        '''
        if self.poller is None:
            return
        events = dict(self.poller.poll(0))

        # The socket with incoming data
        for sock, event in events.items():
            if event != zmq.POLLIN:
                # The event does not refer to incoming data: skip for now
                continue
            # Get the incoming data
            dmsg = self.recv_reqreply()
            if dmsg.conn_id != self.own_id and dmsg.is_server:
                self.server_discovered.emit(dmsg.conn_ip, dmsg.ports)

        self.send_request()


    def send(self, buf):
        self.handle.sendto(buf, 0, (self.broadcast, self.port))

    def receive(self, n):
        return self.handle.recvfrom(n)

    def send_request(self):
        data = msgpack.packb([self.mask|IS_REQUEST])
        self.send(self.own_id + data)

    def send_reply(self, eport, sport):
        data = msgpack.packb([self.mask|IS_REPLY, eport, sport])
        self.send(self.own_id + data)

    def recv_reqreply(self):
        msg, addr = self.receive(IDLEN + 8) # The longest message is IDLEN + 8 bytes
        class DiscoveryReply:
            def __init__(self, msg, addr):
                self.conn_ip = addr[0]
                self.conn_id = msg[:IDLEN]
                data = msgpack.unpackb(msg[IDLEN:])
                self.is_client = bool(data[0] & IS_CLIENT)
                self.is_server = not self.is_client
                self.is_reply = bool(data[0] & IS_REPLY)
                self.is_request = not self.is_reply
                self.ports = data[1:]

            def __repr__(self):
                return 'Discovery {} received from {} {}'.format(
                    'request' if self.is_request else 'reply',
                    'client' if self.is_client else 'server',
                    self.conn_ip)

        return DiscoveryReply(msg, addr)



""" ADS-B model. Implements real-life limitations of ADS-B communication."""
import numpy as np
import bluesky as bs
from bluesky.tools.aero import ft
from bluesky.core import Entity


class ADSB(Entity, replaceable=True):
    """ ADS-B model. Implements real-life limitations of ADS-B communication."""

    def __init__(self):
        super().__init__()
        # From here, define object arrays
        with self.settrafarrays():
            # Most recent broadcast data
            self.lastupdate = np.array([])
            self.lat        = np.array([])
            self.lon        = np.array([])
            self.alt        = np.array([])
            self.trk        = np.array([])
            self.tas        = np.array([])
            self.gs         = np.array([])
            self.vs         = np.array([])

        self.setnoise(False)

    def setnoise(self, n):
        self.transnoise = n
        self.truncated  = n
        self.transerror = [1e-4, 100 * ft]  # [degree,m] standard lat/lon distance, altitude error
        self.trunctime  = 0  # [s]

    def create(self, n=1):
        super().create(n)

        self.lastupdate[-n:] = -self.trunctime * np.random.rand(n)
        self.lat[-n:] = bs.traf.lat[-n:]
        self.lon[-n:] = bs.traf.lon[-n:]
        self.alt[-n:] = bs.traf.alt[-n:]
        self.trk[-n:] = bs.traf.trk[-n:]
        self.tas[-n:] = bs.traf.tas[-n:]
        self.gs[-n:]  = bs.traf.gs[-n:]

    def update(self):
        up = np.where(self.lastupdate + self.trunctime < bs.sim.simt)
        nup = len(up)
        if self.transnoise:
            self.lat[up] = bs.traf.lat[up] + np.random.normal(0, self.transerror[0], nup)
            self.lon[up] = bs.traf.lon[up] + np.random.normal(0, self.transerror[0], nup)
            self.alt[up] = bs.traf.alt[up] + np.random.normal(0, self.transerror[1], nup)
        else:
            self.lat[up] = bs.traf.lat[up]
            self.lon[up] = bs.traf.lon[up]
            self.alt[up] = bs.traf.alt[up]
        self.trk[up] = bs.traf.trk[up]
        self.tas[up] = bs.traf.tas[up]
        self.gs[up]  = bs.traf.gs[up]
        self.vs[up]  = bs.traf.vs[up]
        self.lastupdate[up] = self.lastupdate[up] + self.trunctime



from collections import defaultdict
from typing import Union, Collection
import zmq
import msgpack

import bluesky as bs
from bluesky import stack
from bluesky.core import Entity, Signal
from bluesky.stack.clientstack import process
from bluesky.traffic.trafficproxy import TrafficProxy
from bluesky.network import context as ctx
from bluesky.network.npcodec import encode_ndarray, decode_ndarray
from bluesky.network.subscriber import Subscription
from bluesky.network.common import genid, asbytestr, seqid2idx, MSG_SUBSCRIBE, MSG_UNSUBSCRIBE, GROUPID_NOGROUP, GROUPID_CLIENT, GROUPID_SIM, GROUPID_DEFAULT, IDLEN


# Register settings defaults
bs.settings.set_variable_defaults(recv_port=11000, send_port=11001)


class Client(Entity):
    def __init__(self, group_id=GROUPID_CLIENT):
        super().__init__()
        self.group_id = asbytestr(group_id)
        self.client_id = genid(self.group_id)
        self.act_id = None
        self.acttopics = defaultdict(set)
        self.nodes = set()
        self.servers = set()
        self.discovery = None

        zmqctx = zmq.Context.instance()
        self.sock_recv = zmqctx.socket(zmq.SUB)
        self.sock_send = zmqctx.socket(zmq.XPUB)
        self.poller = zmq.Poller()

        # Tell bluesky that this client will manage the network I/O
        bs.net = self
        # And create a proxy object for easy traffic data access
        bs.traf = TrafficProxy()

        # Subscribe to subscriptions that were already made before constructing
        # this node
        for sub in Subscription.subscriptions.values():
            sub.subscribe_all()

        # Signals
        self.actnode_changed = Signal('actnode-changed')
        self.node_added = Signal('node-added')
        self.node_removed = Signal('node-removed')
        self.server_added = Signal('server-added')
        self.server_removed = Signal('server-removed')

        # If no other object is taking care of this, let this client act as screen object as well
        if not bs.scr:
            bs.scr = self

    def connect(self, hostname=None, recv_port=None, send_port=None, protocol='tcp'):
        ''' Connect client to a server.

            Arguments:
            - hostname: Network name or ip of the server to connect to
            - recv_port: Network port to use for incoming communication
            - send_port: Network port to use for outgoing communication
            - protocol: Network protocol to use
        '''
        conbase = f'{protocol}://{hostname or "localhost"}'
        rcon = conbase + f':{recv_port or bs.settings.recv_port}'
        scon = conbase + f':{send_port or bs.settings.send_port}'
        self.sock_recv.connect(rcon)
        self.sock_send.connect(scon)
        self.poller.register(self.sock_recv, zmq.POLLIN)
        self.poller.register(self.sock_send, zmq.POLLIN)
        # Register this client by subscribing to targeted messages
        self.subscribe('', '', to_group=self.client_id)

    def update(self):
        ''' Client periodic update function.

            Periodically call this function to allow client to receive and process data.
        '''
        self.receive()
        # Process any waiting stacked commands
        process()

    def receive(self, timeout=0):
        ''' Poll for incoming data from Server, and receive if available.
            Arguments:
            timeout: The polling timeout in milliseconds. '''
        try:
            events = dict(self.poller.poll(timeout))

            # The socket with incoming data
            for sock, event in events.items():
                if event != zmq.POLLIN:
                    # The event does not refer to incoming data: skip for now
                    continue
            
                # Receive the message
                ctx.msg = sock.recv_multipart()
                if not ctx.msg:
                    # In the rare case that a message is empty, skip remaning processing
                    continue

                # Regular incoming data
                if sock == self.sock_recv:
                    ctx.topic = ctx.msg[0][IDLEN:-IDLEN].decode()
                    ctx.sender_id = ctx.msg[0][-IDLEN:]
                    pydata = msgpack.unpackb(ctx.msg[1], object_hook=decode_ndarray, raw=False)                    
                    sub = Subscription.subscriptions.get(ctx.topic, None)# or Subscription(ctx.topic, directedonly=True)
                    if sub is None:
                        continue
                    # Unpack dict or list, skip empty string
                    if pydata == '':
                        sub.emit()
                    elif isinstance(pydata, dict):
                        sub.emit(**pydata)
                    elif isinstance(pydata, (list, tuple)):
                        sub.emit(*pydata)
                    else:
                        sub.emit(pydata)
                    ctx.msg = ctx.topic = ctx.sender_id = None

                elif sock == self.sock_send:
                    # This is an (un)subscribe message. If it's an id-only subscription
                    # this is also a registration message
                    if len(ctx.msg[0]) == IDLEN + 1:
                        sender_id = ctx.msg[0][1:]
                        sequence_idx = seqid2idx(sender_id[-1])
                        if sender_id[0] in (GROUPID_SIM, GROUPID_NOGROUP):
                            # This is an initial simulation node subscription
                            if ctx.msg[0][0] == MSG_SUBSCRIBE:
                                if sequence_idx > 0:
                                    self.nodes.add(sender_id)
                                    self.node_added.emit(sender_id)
                                    if not self.act_id:
                                        self.actnode(sender_id)
                                    continue
                                elif sequence_idx == 0:
                                    self.servers.add(sender_id)
                                    self.server_added.emit(sender_id)

                            elif ctx.msg[0][0] == MSG_UNSUBSCRIBE:
                                if sequence_idx > 0:
                                    self.nodes.discard(sender_id)
                                    self.node_removed.emit(sender_id)
                                elif sequence_idx == 0:
                                    self.servers.discard(sender_id)
                                    self.server_removed.emit(sender_id)

        except zmq.ZMQError:
            return False

    def send(self, topic: str, data: Union[str, Collection]='', to_group: str=''):
        btopic = asbytestr(topic)
        btarget = asbytestr(to_group or stack.sender() or self.act_id or '')
        self.sock_send.send_multipart(
            [
                btarget.ljust(IDLEN, b'*') + btopic + self.client_id,
                msgpack.packb(data, default=encode_ndarray, use_bin_type=True)
            ]
        )

    def subscribe(self, topic, from_group=GROUPID_DEFAULT, to_group='', actonly=False):
        ''' Subscribe to a topic.

            Arguments:
            - topic: The name of the topic to subscribe to
            - from_id: The id of the node from which to receive the topic (optional)
            - to_group: The group mask that this topic is sent to (optional)
            - actonly: Set to true if you only want to receive this topic from
              the active node.
        '''
        sub = None
        if topic:
            sub = Subscription(topic)
            sub.actonly = (sub.actonly or actonly)
            actonly = sub.actonly
            if (from_group, to_group) in sub.subs:
                # Subscription already active. Just return Subscription object
                return sub
            sub.subs.add((from_group, to_group))

        self._subscribe(topic, from_group, to_group, actonly)

        # Messages coming in that match this subscription will be emitted using a 
        # subscription signal
        return sub

    def unsubscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        ''' Unsubscribe from a topic.

            Arguments:
            - topic: The name of the stream to unsubscribe from.
            - from_id: When subscribed to data from a specific node: The id of the node
            - to_group: The group mask that this topic is sent to (optional)
        '''
        if topic:
            Subscription(topic).subs.discard((from_group, to_group))
        self._unsubscribe(topic, from_group, to_group)

    def _subscribe(self, topic, from_group=GROUPID_DEFAULT, to_group='', actonly=False):
        if from_group == GROUPID_DEFAULT:
            from_group = GROUPID_SIM
            if actonly:
                self.acttopics[topic].add(to_group)
                if self.act_id is not None:
                    bfrom_group = self.act_id
                else:
                    return
        btopic = asbytestr(topic)
        bfrom_group = asbytestr(from_group)
        bto_group = asbytestr(to_group)

        self.sock_recv.setsockopt(zmq.SUBSCRIBE, bto_group.ljust(IDLEN, b'*') + btopic + bfrom_group)

    def _unsubscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        if from_group == GROUPID_DEFAULT:
            from_group = GROUPID_SIM
        btopic = asbytestr(topic)
        bfrom_group = asbytestr(from_group)
        bto_group = asbytestr(to_group)
        if from_group == GROUPID_DEFAULT and topic in self.acttopics:
            self.acttopics[topic].discard(to_group)
            if self.act_id is not None:
                bfrom_group = self.act_id
            else:
                return
        self.sock_recv.setsockopt(zmq.UNSUBSCRIBE, bto_group.ljust(IDLEN, b'*') + btopic + bfrom_group)

    def actnode(self, newact=None):
        ''' Set the new active node, or return the current active node. '''
        if newact:
            if newact not in self.nodes:
                print('Error selecting active node (unknown node)')
                return None
            # Unsubscribe from previous node, subscribe to new one.
            if newact != self.act_id:
                for topic, groupset in self.acttopics.items():
                    for to_group in groupset:
                        if self.act_id:
                            self._unsubscribe(topic, self.act_id, to_group)
                        self._subscribe(topic, newact, to_group)
                self.act_id = newact
                self.actnode_changed.emit(newact)

        return self.act_id

    def addnodes(self, count=1, *node_ids, server_id=None):
        ''' Tell the specified server to add 'count' nodes. 
        
            If provided, create these nodes with the specified node ids.

            If no server_id is specified, the corresponding server of the
            currently-active node is targeted.
        '''
        self.send(b'ADDNODES', dict(count=count, node_ids=node_ids), server_id or genid(self.act_id[:-1], seqidx=0))



''' BlueSky Stack base data and functions. '''
from itertools import repeat

import bluesky as bs
from bluesky.network import subscriber, context as ctx
from bluesky.stack.cmdparser import command


class Stack:
    ''' Stack static-only namespace. '''

    # Stack data
    current = ''
    cmdstack = []  # The actual stack: Current commands to be processed

    # Scenario details
    scenname = ""  # Currently used scenario name (for reading)
    scentime = []  # Times of the commands from the read scenario file
    scencmd = []  # Commands from the scenario file

    # Current command details
    sender_id = None  # bs net route to sender

    @classmethod
    def reset(cls):
        ''' Reset stack variables. '''
        cls.cmdstack = []
        cls.scenname = ""
        cls.scentime = []
        cls.scencmd = []
        cls.current = ''
        cls.sender_id = None

    @classmethod
    def commands(cls, ext_cmds=None):
        ''' Generator function to iterate over stack commands. '''
        # Return explicitly passed commands if given, otherwise own command stack
        if isinstance(ext_cmds, str):
            # If argument is a single string convert it to an appropriate list
            ext_cmds = zip(ext_cmds.split(';'), repeat(cls.sender_id))

        for cls.current, cls.sender_id in ext_cmds or cls.cmdstack:
            yield cls.current
        # After processing commands, current command and sender id should be reset
        cls.current = ''
        cls.sender_id = None

    @classmethod
    def clear(cls):
        cls.cmdstack.clear()


def checkscen():
    """ Check if commands from the scenario buffer need to be stacked. """
    if Stack.scencmd:
        # Find index of first timestamp exceeding bs.sim.simt
        idx = next((i for i, t in enumerate(
            Stack.scentime) if t > bs.sim.simt), None)
        # Stack all commands before that time, and remove from scenario
        stack(*Stack.scencmd[:idx])
        del Stack.scencmd[:idx]
        del Stack.scentime[:idx]


def stack(*cmdlines, sender_id=None):
    """ Stack one or more commands.
        Stacking multiple commands can be done as multiple comma-separated
        arguments, and/or semicolon-separated within a single string. """
    for cmdline in cmdlines:
        cmdline = cmdline.strip()
        if cmdline:
            for line in cmdline.split(";"):
                Stack.cmdstack.append((line, sender_id))


def forward(*cmdlines, target_id=None):
    ''' Forward one or more stack commands. 

        By default, forward() sends command on to the client if this stack is
        running sim-side, and vice-versa. Instead, a target id can be specified
        to specifically send the stack command(s) to this node.
    
        Multiple commands can be specified as multiple comma-separated
        arguments, and/or semicolon-separated within a single string.
    '''
    if target_id is not None or Stack.sender_id is None:
        # Only forward if this command originated here
        bs.net.send(b'STACK', ';'.join(cmdlines) if cmdlines else Stack.current, target_id)


def sender():
    """ Return the sender of the currently executed stack command.
        If there is no sender id (e.g., when the command originates
        from a scenario file), None is returned. """
    return Stack.sender_id


def get_scenname():
    """ Return the name of the current scenario.
        This is either the name defined by the SCEN command,
        or otherwise the filename of the scenario. """
    return Stack.scenname


def get_scendata():
    """ Return the scenario data that was loaded from a scenario file. """
    return Stack.scentime, Stack.scencmd


def set_scendata(newtime, newcmd):
    """ Set the scenario data. This is used by the batch logic. """
    Stack.scentime = newtime
    Stack.scencmd = newcmd


# Register subscriber for stack commands coming from the network
@subscriber(topic='STACK')
def on_stack_received(data):
    """ Add stack commands coming from the network to the stack. """
    stack(data, sender_id=ctx.sender_id)


@command
def stackat(targetid, cmdline: 'string'):
    """ Send a stack command to a specific node (client or simulation) or group. """
    forward(cmdline, target_id=targetid)


import os
import inspect


class FuncObjectMeta(type):
    def __call__(cls, func, *args, **kwargs):
        # Retrieve preexisting FuncObject, if it exists
        fobj = getattr(inspect.unwrap(func), '__func_object__', None)
        # Don't return the preexisting FuncObject if func is a method bound to an instance that is
        # not the first instance of its type.
        if fobj is not None:
            if not inspect.ismethod(fobj.func) or func is fobj.func:
                return fobj
        return super().__call__(func, *args, **kwargs)



class FuncObject(metaclass=FuncObjectMeta):
    ''' Function reference object that is automatically updated
        on implementation selection for replaceables, and on creation of
        instances.
    '''
    __slots__ = ['func', 'callback']

    def __init__(self, func) -> None:
        ifunc = inspect.unwrap(func, stop=lambda f:not isinstance(func, (staticmethod, classmethod)))
        self.update(ifunc)
        ufunc = inspect.unwrap(func)
        setattr(getattr(ufunc, '__func__', ufunc), '__func_object__', self)

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def __repr__(self) -> str:
        return repr(self.func)

    def __eq__(self, value) -> bool:
        return self is value or \
            self.func == inspect.unwrap(value)

    def notimplemented(self, *args, **kwargs):
        pass

    def update(self, func):
        self.func = func
        self.callback = func if self.valid else self.notimplemented

    def info(self):
        msg = ''
        if self.func.__name__ == '<lambda>':
            msg += 'Anonymous (lambda) function, implemented in '
        else:
            msg += f'Function {self.func.__name__}(), implemented in '
        if hasattr(self.func, '__code__'):
            fname = self.func.__code__.co_filename
            fname_stripped = fname.replace(os.getcwd(), '').lstrip('/')
            firstline = self.func.__code__.co_firstlineno
            msg += f'<a href="file://{fname}">{fname_stripped} on line {firstline}</a>'
        else:
            msg += f'module {self.func.__module__}'

        return msg

    @property
    def __wrapped__(self):
        return self.func

    @property
    def __name__(self):
        return self.func.__name__

    @property
    def valid(self):
        if self.func is None:
            return False
        spec = inspect.signature(self.func)
        # Check if this is an unbound class/instance method
        return spec.parameters.get('self') is None and \
            spec.parameters.get('cls') is None



''' Eby ConflictResolution implementation plugin. '''
import numpy as np
from bluesky.tools.aero import vtas2eas
from bluesky.traffic.asas import ConflictResolution


def init_plugin():

    # Addtional initilisation code

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'EBY',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim'
    }

    return config


class Eby(ConflictResolution):
    def resolve(self, conf, ownship, intruder):
        ''' Resolve all current conflicts '''
        # required change in velocity
        dv = np.zeros((ownship.ntraf, 3))

        for ((ac1, ac2), qdr, dist, tcpa, tLOS) in zip(conf.confpairs, conf.qdr, conf.dist, conf.tcpa, conf.tLOS):
            idx1 = ownship.id.index(ac1)
            idx2 = intruder.id.index(ac2)
            if idx1 > -1 and idx2 > -1:
                dv_eby = self.Eby_straight(
                    ownship, intruder, conf, qdr, dist, tcpa, tLOS, idx1, idx2)
                dv[idx1] -= dv_eby
                dv[idx2] += dv_eby

        # now we have the change in speed vector for each aircraft.
        dv=np.transpose(dv)
        # the old speed vector, cartesian coordinates
        trkrad = np.radians(ownship.trk)
        v = np.array([np.sin(trkrad) * ownship.tas,\
                      np.cos(trkrad)*ownship.tas,\
                      ownship.vs])
        # the new speed vector
        newv = dv + v

        # the new speed vector in polar coordinates
        newtrack = (np.arctan2(newv[0,:], newv[1,:]) * 180 / np.pi) % 360
        newgs = np.sqrt(newv[0,:] ** 2 + newv[1,:] ** 2)
        neweas = vtas2eas(newgs, ownship.alt)

        # Cap the velocity
        neweascapped = np.maximum(ownship.perf.vmin, np.minimum(ownship.perf.vmax, neweas))

        return newtrack, neweascapped, newv[2, :], np.sign(newv[2, :]) * 1e5

    def Eby_straight(self, ownship, intruder, conf, qdr, dist, tcpa, tLOS, idx1, idx2):
        ''' 
            Resolution: Eby method assuming aircraft move straight forward,
            solving algebraically, only horizontally.
        '''
        # from degrees to radians
        qdr  = np.radians(qdr)
        # relative position vector
        d    = np.array([np.sin(qdr) * dist, \
                         np.cos(qdr) * dist, \
                         intruder.alt[idx2] - ownship.alt[idx1]])

        # find track in radians
        t1 = np.radians(ownship.trk[idx1])
        t2 = np.radians(intruder.trk[idx2])

        # write velocities as vectors and find relative velocity vector
        v1 = np.array([np.sin(t1) * ownship.tas[idx1], np.cos(t1) * ownship.tas[idx1], ownship.vs[idx1]])
        v2 = np.array([np.sin(t2) * intruder.tas[idx2], np.cos(t2) * intruder.tas[idx2], intruder.vs[idx2]])
        v = np.array(v2 - v1)
        # bear in mind: the definition of vr (relative velocity) is opposite to
        # the velocity vector in the LOS_nominal method, this just has consequences
        # for the derivation of tstar following Eby method, not more
        """
        intrusion vector:
        i(t)=self.hsep-d(t)
        d(t)=sqrt((d[0]+v[0]*t)**2+(d[1]+v[1]*t)**2)
        find max(i(t)/t)
        -write the equation out
        -take derivative, set to zero
        -simplify, take square of everything so the sqrt disappears (creates two solutions)
        -write to the form a*t**2 + b*t + c = 0
        -Solve using the quadratic formula
        """
        # These terms are used to construct a,b,c of the quadratic formula
        R2 = (conf.rpz * self.resofach) ** 2 # in meters
        d2 = np.dot(d, d) # distance vector length squared
        v2 = np.dot(v, v) # velocity vector length squared
        dv = np.dot(d, v) # dot product of distance and velocity

        # Solving the quadratic formula
        a = R2 * v2 - dv **2
        b = 2 * dv * (R2 - d2)
        c = R2 * d2 - d2 ** 2
        discrim = b ** 2 - 4 * a * c

        if discrim < 0: # if the discriminant is negative, we're done as taking the square root will result in an error
            discrim = 0
        time1 = (-b + np.sqrt(discrim)) / (2 * a)
        time2 = (-b - np.sqrt(discrim)) / (2 * a)

        #time when the size of the conflict is largest relative to time to solve
        tstar = min(abs(time1), abs(time2))

        #find drel and absolute distance at tstar
        drelstar = d + v * tstar
        dstarabs = np.linalg.norm(drelstar)
        #exception: if the two aircraft are on exact collision course
        #(passing eachother within 10 meter), change drelstar
        exactcourse = 10 #10 meter
        dif = exactcourse - dstarabs
        if dif > 0:
            vperp = np.array([-v[1], v[0], 0]) #rotate velocity 90 degrees in horizontal plane
            drelstar += dif * vperp / np.linalg.norm(vperp) #normalize to 10 m and add to drelstar
            dstarabs = np.linalg.norm(drelstar)

        #intrusion at tstar
        i = (conf.rpz * self.resofach) - dstarabs

        #desired change in the plane's speed vector:
        dv = i * drelstar / (dstarabs * tstar)
        return dv



''' BlueSky navdata OpenGL visualisation object. '''
import numpy as np
import bluesky as bs
from bluesky.core import Signal
from bluesky.network import context as ctx
from bluesky.network.subscriber import subscriber
from bluesky.network.sharedstate import ActData
from bluesky.stack import command
from bluesky.ui.qtgl import glhelpers as glh
from bluesky import settings
from bluesky.ui.loadvisuals import load_aptsurface
from bluesky.ui import palette


# Register settings defaults
settings.set_variable_defaults(
    gfx_path='graphics',
    text_size=13, apt_size=10,
    wpt_size=10)

palette.set_default_colours(
    aptlabel=(220, 250, 255),
    aptsymbol=(148, 178, 235),
    pavement=(160, 160, 160),
    runways=(100, 100, 100),
    taxiways=(100, 100, 100),
    thresholds=(255, 255, 255),
    wptlabel=(220, 250, 255),
    wptsymbol=(148, 178, 235)
)

# Static defines
CUSTWP_SIZE = 1000


class Navdata(glh.RenderObject, layer=-10):
    ''' Navdata OpenGL object. '''

    # Per remote node attributes
    show_wpt: ActData[int] = ActData(1)
    show_apt: ActData[int] = ActData(1)
    pan: ActData[list[float]] = ActData([0.0, 0.0], group='panzoom')
    zoom: ActData[float] = ActData(1.0, group='panzoom')

    @command
    def showwpt(self, flag:int=None):
        ''' Toggle drawing of waypoints. '''
        # TODO: add to SWRAD
        # flag == 'VOR' or flag == 'WPT' or flag == 'WP' or flag == 'NAV':
        # Cycle waypoint visualisation through detail level 0,1,2
        if flag is None:
            self.show_wpt = (self.show_wpt + 1) % 3

        # Or use the argument if it is an integer
        else:
            self.show_wpt = min(2,max(0,flag))

    @command
    def showapt(self, flag:int=None):
        ''' Toggle drawing of waypoints. '''
        # Cycle waypoint visualisation through detail level 0,1,2
        # Airport: 0 = None, 1 = Large, 2= All
        if flag is None:
            self.show_apt = (self.show_apt + 1) % 3

        # Or use the argument if it is an integer
        else:
            self.show_apt = min(2,max(0,flag))

    def __init__(self, parent=None):
        super().__init__(parent)
        self.custwplblbuf = glh.GLBuffer()

        # ------- Airport graphics -----------------------
        self.runways = glh.VertexArrayObject(glh.gl.GL_TRIANGLES)
        self.thresholds = glh.VertexArrayObject(glh.gl.GL_TRIANGLES)
        self.taxiways = glh.VertexArrayObject(glh.gl.GL_TRIANGLES)
        self.pavement = glh.VertexArrayObject(glh.gl.GL_TRIANGLES)

        self.apt_ctrlat = []
        self.apt_ctrlon = []
        self.apt_indices = []
        self.apt_inrange = np.array([])

        self.nwaypoints = 0
        self.nnavaids = 0
        self.ncustwpts = 0
        self.waypoints = glh.VertexArrayObject(glh.gl.GL_LINE_LOOP)
        self.wptlabels = glh.Text(settings.text_size, (5, 1))
        self.customwp = glh.VertexArrayObject(glh.gl.GL_LINE_LOOP)
        self.customwplbl = glh.Text(settings.text_size, (10, 1))

        self.nairports = []
        self.airports = glh.VertexArrayObject(glh.gl.GL_LINE_LOOP)
        self.aptlabels = glh.Text(settings.text_size, (4, 1))

        self.vbuf_asphalt, self.vbuf_concrete, self.vbuf_runways, self.vbuf_rwythr, \
            self.apt_ctrlat, self.apt_ctrlon, self.apt_indices = load_aptsurface()

        Signal('state-changed.panzoom').connect(self.panzoom)

    def panzoom(self, data, finished=True):
        if ctx.action is None or ctx.action == ctx.action.Reset:
            return
        # TODO: need good way to set defaults!!
        if not hasattr(data, 'zoom'):
            data.zoom = 1.0
        if not hasattr(data, 'pan'):
            data.pan = [0.0, 0.0]
        if data.zoom >= 1.0:
            # Airports may be visible when zoom > 1: in this case, update the list of indicates
            # of airports that need to be drawn
            ll_range = max(1.5 / data.zoom, 1.0)
            indices = np.logical_and(np.abs(self.apt_ctrlat - data.pan[0])
                                    <= ll_range, np.abs(self.apt_ctrlon - data.pan[1]) <= ll_range)
            self.apt_inrange = self.apt_indices[indices]
        else:
            self.apt_inrange = np.array([])
    
    @subscriber
    def defwpt(self, data):
        ''' Receive custom waypoint data and add to visualisation. '''
        if not data.valid():
            self.ncustwpts = 0
            return

        self.ncustwpts = len(data.custwplbl)
        self.customwp.update(lat=np.array(data.custwplat, dtype=np.float32),
                                lon=np.array(data.custwplon, dtype=np.float32))
        lbl = [n[:10].ljust(10) for n in data.custwplbl]
        self.custwplblbuf.update(
            np.array(lbl, dtype=np.bytes_))

    def create(self):
        apt_size = settings.apt_size
        wpt_size = settings.wpt_size
        # self.custwplblbuf.create(CUSTWP_SIZE * 10, usage=glh.gl.GL_STATIC_DRAW)
        self.custwplblbuf.create(CUSTWP_SIZE * 10, usage=glh.GLBuffer.UsagePattern.StaticDraw)

        # Load vertex data
        
        self.runways.create(vertex=self.vbuf_runways, color=palette.runways)
        self.thresholds.create(vertex=self.vbuf_rwythr, color=palette.thresholds)
        self.taxiways.create(vertex=self.vbuf_asphalt, color=palette.taxiways)
        self.pavement.create(vertex=self.vbuf_concrete, color=palette.pavement)
        del self.vbuf_asphalt, self.vbuf_concrete, self.vbuf_runways, self.vbuf_rwythr

        # ------- Waypoints ------------------------------
        wptvertices = np.array([(0.0, 0.5 * wpt_size), (-0.5 * wpt_size, -0.5 * wpt_size),
                                (0.5 * wpt_size, -0.5 * wpt_size)], dtype=np.float32)  # a triangle
        self.nwaypoints = len(bs.navdb.wplat)
        self.waypoints.create(vertex=wptvertices, color=palette.wptsymbol,
                              n_instances=self.nwaypoints)
        # Sort based on id string length
        llid = sorted(zip(bs.navdb.wpid, bs.navdb.wplat,
                          bs.navdb.wplon), key=lambda i: len(i[0]) > 3)
        wpidlst, wplat, wplon = zip(*llid)
        self.waypoints.set_attribs(lat=np.array(wplat, dtype=np.float32), lon=np.array(
            wplon, dtype=np.float32), instance_divisor=1)
        wptids = ''
        self.nnavaids = 0
        for wptid in wpidlst:
            if len(wptid) <= 3:
                self.nnavaids += 1
            wptids += wptid[:5].ljust(5)
        npwpids = np.array(wptids, dtype=np.bytes_)
        self.wptlabels.create(npwpids, self.waypoints.lat,
                              self.waypoints.lon, palette.wptlabel,
                              (wpt_size, 0.5 * wpt_size), instanced=True)

        self.customwp.create(vertex=self.waypoints.vertex, vertex_count=3, color=palette.wptsymbol)
        self.customwp.set_attribs(
            lat=CUSTWP_SIZE * 4, lon=CUSTWP_SIZE * 4, instance_divisor=1)
        self.customwplbl.create(self.custwplblbuf,
                                self.customwp.lat, self.customwp.lon,
                                palette.wptlabel, (wpt_size, 0.5 * wpt_size),
                                instanced=True)

        # ------- Airports -------------------------------
        aptvertices = np.array([(-0.5 * apt_size, -0.5 * apt_size),
                                (0.5 * apt_size, -0.5 * apt_size),
                                (0.5 * apt_size, 0.5 * apt_size),
                                (-0.5 * apt_size, 0.5 * apt_size)], dtype=np.float32)  # a square
        self.nairports = len(bs.navdb.aptlat)
        self.airports.create(vertex=aptvertices, color=palette.aptsymbol,
                             n_instances=self.nairports)
        indices = bs.navdb.aptype.argsort()
        aplat = np.array(bs.navdb.aptlat[indices], dtype=np.float32)
        aplon = np.array(bs.navdb.aptlon[indices], dtype=np.float32)
        aptypes = bs.navdb.aptype[indices]
        apnames = np.array(bs.navdb.aptid)
        apnames = apnames[indices]
        # The number of large, large+med, and large+med+small airports
        self.nairports = [aptypes.searchsorted(
            2), aptypes.searchsorted(3), self.nairports]

        self.airports.set_attribs(lat=aplat, lon=aplon, instance_divisor=1)
        aptids = ''
        for aptid in apnames:
            aptids += aptid.ljust(4)
        self.aptlabels.create(np.array(aptids, dtype=np.bytes_),
                              self.airports.lat, self.airports.lon,
                              palette.aptlabel, (apt_size, 0.5 * apt_size),
                              instanced=True)

    def draw(self):
        # Send the (possibly) updated global uniforms to the buffer
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_LATLON)

        self.shaderset.enable_wrap(False)

        # --- DRAW AIRPORT DETAILS (RUNWAYS, TAXIWAYS, PAVEMENTS) -------------
        self.runways.draw()
        self.thresholds.draw()

        if self.zoom >= 1.0:
            for idx in self.apt_inrange:
                self.taxiways.draw(first_vertex=idx[0], vertex_count=idx[1])
                self.pavement.draw(first_vertex=idx[2], vertex_count=idx[3])

        self.shaderset.enable_wrap(True)
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_SCREEN)

        if self.zoom >= 0.5 and self.show_apt == 1 or self.show_apt == 2:
            nairports = self.nairports[2]
        elif self.zoom >= 0.25 and self.show_apt == 1 or self.show_apt == 3:
            nairports = self.nairports[1]
        else:
            nairports = self.nairports[0]

        if self.zoom >= 3 and self.show_wpt == 1 or self.show_wpt == 2:
            nwaypoints = self.nwaypoints
        else:
            nwaypoints = self.nnavaids

        # Draw waypoint symbols
        if self.show_wpt:
            self.waypoints.draw(n_instances=nwaypoints)
            if self.ncustwpts > 0:
                self.customwp.draw(n_instances=self.ncustwpts)

        # Draw airport symbols
        if self.show_apt:
            self.airports.draw(n_instances=nairports)

        # Draw wpt/apt labels
        if self.show_apt:
            self.aptlabels.draw(n_instances=nairports)
        if self.show_wpt:
            self.wptlabels.draw(n_instances=nwaypoints)
            if self.ncustwpts > 0:
                self.customwplbl.draw(n_instances=self.ncustwpts)



"""
Aircraft Performance Modeling tools

Modules:
     esf (performance parameters): Energy Share Factor for climb/descent performance
     phases (performance parameters): define flight phases
     vmin (min speeds, phase)       : summarize minimum speeds
     limits (performance parameters): ensure aircraft remains within flight envelope

Created by  : Isabel Metz
Date        : February 2015

Modification  :
By            :
Date          :

"""

import numpy as np
from bluesky.tools.aero import kts, ft, gamma, gamma1, gamma2, R, beta, g0, \
    vmach2cas


PHASE = {"None":0,
         "TO"  :1, # Take-off
         "IC"  :2, # Initial climb
         "CR"  :3, # Cruise
         "AP"  :4, # Approach
         "LD"  :5, # Landing
         "GD"  :6, # Ground
         "to"  :1,
         "ic"  :2,
         "cr"  :3,  # and lower case to be sure
         "ap"  :4,
         "ld"  :5,
         "gd"  :6}

#------------------------------------------------------------------------------
#
# FLIGHT PHASES
# based on the BADA 3.12 User Manual, chapter 3.5, p. 19, adapted
#------------------------------------------------------------------------------


def phases(alt, gs, delalt, cas, vmto, vmic, vmap,
           vmcr, vmld, bank, bphase, swhdgsel, bada):
    # flight phases: TO (1), IC (2), CR (3), AP(4), LD(5), GD (6)
    #--> no holding phase yet
    #-------------------------------------------------
    # phase TO[1]: alt<400 and vs>0
    Talt = np.array(alt < (400. * ft))
    Tspd = np.array(gs > (30. * kts))
    Tvs  = np.array(delalt >= 0.) * 1.0

    to   = Talt * Tspd * Tvs * 1

    #-------------------------------------------------
    # phase IC[2]: 400<alt<2000, vs>0
    Ialt = np.array((alt >= (400. * ft)) & (alt < (2000. * ft)))
    Ivs  = np.array(delalt > 0.)

    ic   = Ialt*Ivs*2


    #-------------------------------------------------
    #phase CR[3]: in climb above 2000ft, in descent
    # above 8000ft and below 8000ft if V>=Vmincr + 10kts

    # a. climb
    Caalt = np.array(alt >= (2000. * ft))
    Cavs  = np.array(delalt >= 0.)
    cra   = np.logical_and.reduce([Caalt, Cavs]) * 1

    #b. above 8000ft
    crb   = np.array(alt > (8000. * ft)) * 1

    #c. descent
    Ccalt = np.array(alt <= (8000. * ft))
    Ccvs  = np.array(delalt <= 0.)
    Ccspd = np.array(cas >= (vmcr + 10. * kts))
    crc   = np.logical_and.reduce([Ccalt, Ccvs, Ccspd]) * 1

    # merge climb and descent phase
    cr = np.maximum(cra, np.maximum(crb, crc)) * 3

    #-------------------------------------------------
    # phase AP[4]

    #a. alt<8000, Speed between Vmcr+10 and Vmapp+10, v<>0
    #altitude check
    Aaalt = np.array((alt > ft) & (alt <= (8000. * ft)))
    Aaspd = np.array(cas < (vmcr + 10. * kts))
    Aavs  = np.array(delalt <= 0.)
    apa   = np.logical_and.reduce([Aaalt, Aaspd , Aavs]) * 1
    #b. alt<3000ft, Vmcr+10>V>Vmap+10
    Abalt = np.array((alt > ft) & (alt <= (3000. * ft)))
    if bada:
        Abspd = np.array((cas >= (vmap + 10. * kts)) & (cas < (vmcr + 10. * kts)))
    else:
        Abspd = np.array(cas >= (vmap + 10. * kts))
    Abvs = np.array(delalt <= 0.)
    apb  = np.logical_and.reduce([Abalt, Abspd , Abvs]) * 1

    # merge a. and b.
    ap   = np.maximum.reduce([apa, apb]) * 4

    #-------------------------------------------------
    # phase LD[5]: alt<3000, Speed below vmap (i.e. minimum approach speed) + 10kt
    Lalt = np.array(alt <= (3000.0 * ft))
    if bada:
        Lspd = np.array(cas < (vmap + 10.0 * kts))
    else:
        Lspd = np.array(gs >= (30.0 * kts))
    Lvs = np.array(delalt <= 0.0)
    ld  = np.logical_and.reduce([Lalt, Lspd, Lvs]) * 5

    #-------------------------------------------------
    # phase GND: alt < 1 ft, i.e. as soon as on ground
    gd = np.array(alt <= ft)* 6

    #-------------------------------------------------
    # combine all phases
    phase = np.maximum.reduce([to, ic, ap, ld, cr, gd])

    to2 = np.where(phase == 1)
    ic2 = np.where(phase == 2)
    cr2 = np.where(phase == 3)
    ap2 = np.where(phase == 4)
    ld2 = np.where(phase == 5)
    gd2 = np.where(phase == 6)

    # assign aircraft to their nominal bank angle per phase
    bank[to2] = bphase[0]
    bank[ic2] = bphase[1]
    bank[cr2] = bphase[2]
    bank[ap2] = bphase[3]
    bank[ld2] = bphase[4]
    # to be refined! find value that comes closest to 1 if tan or cos
    bank[gd2] = bphase[5]

    # not turning aircraft do not have a bank angle.
    #swhdgsel == True: Aircraft is turning
    noturn = np.array(swhdgsel) * 100.0
    bank   = np.minimum(noturn, bank)

    return (phase, bank)


#------------------------------------------------------------------------------
#
# ENERGY SHARE FACTOR
# (BADA User Manual 3.12, p.15)
#
#-----------------------------------------------------------------------------
def esf(alt, M, climb, descent, delspd, selmach):

    # test for acceleration / deceleration
    cspd  = np.array((delspd <= 0.001) & (delspd >= -0.001))
    # accelerating or decelerating
    acc   = np.array(delspd > 0.001)
    dec   = np.array(delspd < -0.001)

    # tropopause
    abtp  = np.array(alt > 11000.0)
    beltp = np.logical_not(abtp)

    selcas = np.logical_not(selmach)

    # constant Mach/CAS
    # case a: constant MA above TP
    efa   = np.logical_and.reduce([cspd, selmach, abtp]) * 1

    # case b: constant MA below TP (at the moment just ISA: tISA = 1)
    # tISA = (self.temp-self.dtemp)/self.temp
    efb   = 1.0 / ((1.0 + ((gamma * R * beta) / (2.0 * g0)) * M**2)) \
        * np.logical_and.reduce([cspd, selmach, beltp]) * 1

    # case c: constant CAS below TP (at the moment just ISA: tISA = 1)
    efc = 1.0 / (1.0 + (((gamma * R * beta) / (2.0 * g0)) * (M**2)) +
        ((1.0 + gamma1 * (M**2))**(-1.0 / (gamma - 1.0))) *
        (((1.0 + gamma1 * (M**2))**gamma2) - 1)) * \
        np.logical_and.reduce([cspd, selcas, beltp]) * 1

    #case d: constant CAS above TP
    efd = 1.0 / (1.0 + ((1.0 + gamma1 * (M**2))**(-1.0 / (gamma - 1.0))) *
        (((1.0 + gamma1 * (M**2))**gamma2) - 1.0)) * \
        np.logical_and.reduce([cspd, selcas, abtp]) * 1

    #case e: acceleration in climb
    efe    = 0.3 * np.logical_and.reduce([acc, climb])

    #case f: deceleration in descent
    eff    = 0.3 * np.logical_and.reduce([dec, descent])

    #case g: deceleration in climb
    efg    = 1.7 * np.logical_and.reduce([dec, climb])

    #case h: acceleration in descent
    efh    = 1.7 * np.logical_and.reduce([acc, descent])

    # combine cases
    ef = np.maximum.reduce([efa, efb, efc, efd, efe, eff, efg, efh])

    # ESF of non-climbing/descending aircraft is zero which
    # leads to an error. Therefore, ESF for non-climbing aircraft is 1
    return np.maximum(ef, np.array(ef == 0) * 1)


#------------------------------------------------------------------------------
#
# CALCULATE LIMITS
#
#------------------------------------------------------------------------------
def calclimits(desspd, gs, to_spd, vmin, vmo, mmo, M, alt, hmaxact,
           desalt, desvs, maxthr, Thr, D, tas, mass, ESF, phase):

    # minimum CAS - below crossover (we do not check for minimum Mach)
    limspd      = np.where((desspd < vmin), vmin, -999.)

    # in traf, we will check for min and max spd, hence a flag is required
    limspd_flag = np.where((desspd < vmin), True, False)

    # maximum CAS: below crossover and above crossover
    limspd      = np.where((desspd > vmo), vmo, limspd )
    limspd_flag = np.where((desspd > vmo), True, limspd_flag)

    # maximum Mach
    limspd      = np.where((M > mmo), vmach2cas(mmo, alt), limspd)
    limspd_flag = np.where((M > mmo), True, limspd_flag)

    # remove non-needed limits
    limspd_flag = np.where((np.abs(desspd-limspd) <0.1), False, limspd_flag)
    limspd      = np.where((limspd_flag==False), -999.,limspd)

    # set altitude to max. possible altitude if alt>Hmax
    limalt = np.where((desalt>hmaxact), hmaxact -1.0, -999.)
    limalt_flag = np.where((desalt>hmaxact), True, False)

    # remove non-needed limits
    limalt = np.where((np.abs(desalt-hmaxact)<0.1), -999., limalt)
    limalt_flag = np.where((np.abs(desalt-hmaxact)<0.1), False, limalt_flag)

    # thrust and vertical speed
    Thr_corrected   = np.where((Thr > maxthr-1.0), maxthr-1., Thr)
    limvs = np.where((Thr >maxthr-1.0), ((Thr_corrected - D) * tas) / (mass * g0)* ESF, -9999.0)
    limvs_flag = np.where(limvs > -9999.0 , True, False)

    # aircraft can only take-off as soon as their speed is above v_rotate
    # True means that current speed is below rotation speed
    # limit vertical speed is thrust limited and thus should only be
    # applied for aircraft that are climbing
    limvs       = np.where((desvs > 0.) & (gs<to_spd) & (phase == 6), 0.0, limvs)
    limvs_flag  = np.where ((desvs > 0.) & (gs<to_spd) & (phase == 6), True, limvs_flag)


    # remove takeoff limit
    limvs      = np.where ((np.abs(to_spd - gs) < 0.1) & ((phase == 6) | (phase == 1)), -9999.,  limvs)
    limvs_flag = np.where((np.abs(to_spd - gs) < 0.1) & ((phase == 6) | (phase == 1)), True, limvs_flag)



    # remove non-needed limits
    Thr        = np.where((maxthr-Thr< 2.), -9999., Thr)
    limvs      = np.where((maxthr-Thr< 2.), -9999., limvs)
    limvs_flag = np.where((limvs< -999.), False, limvs_flag)


    return limspd, limspd_flag, limalt, limalt_flag, limvs, limvs_flag



''' BlueSky OpenGL radar view. '''
from ctypes import c_float, c_int, Structure
import numpy as np

from bluesky.tools import areafilter
from PyQt6.QtCore import Qt, QEvent

import bluesky as bs
from bluesky.core import Signal
import bluesky.network.context as ctx
import bluesky.network.sharedstate as ss
from bluesky.ui.qtgl import glhelpers as glh

from .gltraffic import Traffic
from .glmap import Map
from .glnavdata import Navdata
from .glpoly import Poly
from .gltiledmap import TiledMap


# Register settings defaults
bs.settings.set_variable_defaults(gfx_path='graphics')


class RadarShaders(glh.ShaderSet):
    ''' Shaderset for the radar view. '''
    # Vertex type enum: Individual vertices can correspond to lat/lon coordinates, screen coordinates, or meters.
    VERTEX_IS_LATLON, VERTEX_IS_METERS, VERTEX_IS_SCREEN = list(range(3))
    def __init__(self, parent):
        super().__init__(parent)

        class GlobalData(Structure):
            _fields_ = [("wrapdir", c_int), ("wraplon", c_float), ("panlat", c_float), ("panlon", c_float),
                        ("zoom", c_float), ("screen_width", c_int), ("screen_height", c_int), ("vertex_scale_type", c_int),
                        ("screen_pixel_ratio", c_float)]
        self.data = GlobalData()

    def create(self):
        super().create()
        shaderpath = (bs.resource(bs.settings.gfx_path) / 'shaders').as_posix()
        self.set_shader_path(shaderpath)
        # Load all shaders for this shader set
        self.load_shader('normal', 'radarwidget-normal.vert',
                         'radarwidget-color.frag')
        self.load_shader('textured', 'radarwidget-normal.vert',
                         'radarwidget-texture.frag')
        self.load_shader('tiled', 'radarwidget-normal.vert',
                         'radarwidget-tiled.frag')
        self.load_shader('text', 'radarwidget-text.vert',
                         'radarwidget-text.frag')
        self.load_shader('ssd', 'ssd.vert', 'ssd.frag', 'ssd.geom')

    def set_wrap(self, wraplon, wrapdir):
        self.data.wrapdir = wrapdir
        self.data.wraplon = wraplon

    def set_pan_and_zoom(self, panlat, panlon, zoom):
        self.data.panlat = panlat
        self.data.panlon = panlon
        self.data.zoom = zoom

    def set_pixel_ratio(self, pxratio):
        self.data.screen_pixel_ratio = pxratio

    def set_win_width_height(self, w, h):
        self.data.screen_width = w
        self.data.screen_height = h

    def enable_wrap(self, flag=True):
        if not flag:
            wrapdir = self.data.wrapdir
            self.data.wrapdir = 0
            self.update_ubo('global_data', self.data, 0, 4)
            self.data.wrapdir = wrapdir
        else:
            self.update_ubo('global_data', self.data, 0, 4)

    def set_vertex_scale_type(self, vertex_scale_type):
        self.data.vertex_scale_type = vertex_scale_type
        self.update_ubo('global_data', self.data)


class RadarWidget(glh.RenderWidget):
    ''' The BlueSky radar view. '''

    # Per-remote attributes
    pan: ss.ActData[list[float]] = ss.ActData([0.0, 0.0], group='panzoom')
    zoom: ss.ActData[float] = ss.ActData(1.0, group='panzoom')

    def __init__(self, parent=None):
        super().__init__(parent)
        self.prevwidth = self.prevheight = 600
        self.pxratio = 1
        self.ar = 1.0
        self.flat_earth = 1.0
        self.wraplon = int(-999)
        self.wrapdir = int(0)
        self.initialized = False

        self.panzoomchanged = False
        self.mousedragged = False
        self.mousepos = (0, 0)
        self.prevmousepos = (0, 0)

        self.shaderset = RadarShaders(self)
        self.set_shaderset(self.shaderset)

        # Add default objects
        self.addobject(Map(parent=self))
        self.addobject(Traffic(parent=self))
        self.addobject(Navdata(parent=self))
        self.addobject(Poly(parent=self))

        self.setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, True)
        self.grabGesture(Qt.GestureType.PanGesture)
        self.grabGesture(Qt.GestureType.PinchGesture)
        # self.grabGesture(Qt.SwipeGesture)
        self.setMouseTracking(True)

        # Signals and slots
        self.mouse_event = Signal('radarmouse')
        self.radarclick_event = Signal('radarclick')
        self.panzoom_event = Signal('state-changed.panzoom')
        self.panzoom_event.connect(self.on_panzoom)

    def initializeGL(self):
        """Initialize OpenGL, VBOs, upload data on the GPU, etc."""
        super().initializeGL()

        # Set initial values for the global uniforms
        self.shaderset.set_wrap(self.wraplon, self.wrapdir)
        self.shaderset.set_pan_and_zoom(self.pan[0], self.pan[1], self.zoom)

        # background color
        glh.gl.glClearColor(0.7, 0.7, 0.7, 0)
        glh.gl.glEnable(glh.gl.GL_BLEND)
        glh.gl.glBlendFunc(glh.gl.GL_SRC_ALPHA, glh.gl.GL_ONE_MINUS_SRC_ALPHA)

        self.initialized = True


    def resizeGL(self, width, height):
        """ Called upon window resizing: reinitialize the viewport. """
        # update the window size

        # Calculate zoom so that the window resize doesn't affect the scale, but only enlarges or shrinks the view
        zoom = float(self.prevwidth) / float(width)
        origin = (width / 2, height / 2)

        # Update width, height, and aspect ratio
        self.prevwidth, self.prevheight = width, height
        self.ar = float(width) / max(1, float(height))
        self.pxratio = self.devicePixelRatio()
        self.shaderset.set_pixel_ratio(self.pxratio)
        self.shaderset.set_win_width_height(width, height)
        # Update zoom
        self.setpanzoom(zoom=zoom, origin=origin, absolute=False)

    def pixelCoordsToGLxy(self, x, y):
        """Convert screen pixel coordinates to GL projection coordinates (x, y range -1 -- 1)
        """
        # GL coordinates (x, y range -1 -- 1)
        glx = (float(2.0 * x) / self.prevwidth - 1.0)
        gly = -(float(2.0 * y) / self.prevheight - 1.0)
        return glx, gly

    def pixelCoordsToLatLon(self, x, y):
        """Convert screen pixel coordinates to lat/lon coordinates
        """
        glx, gly = self.pixelCoordsToGLxy(x, y)

        # glxy   = zoom * (latlon - pan)
        # latlon = pan + glxy / zoom
        lat = self.pan[0] + gly / (self.zoom * self.ar)
        lon = self.pan[1] + glx / (self.zoom * self.flat_earth)
        return lat, lon

    def viewportlatlon(self):
        ''' Return the viewport bounds in lat/lon coordinates. '''
        return (self.pan[0] + 1.0 / (self.zoom * self.ar),
                self.pan[1] - 1.0 / (self.zoom * self.flat_earth),
                self.pan[0] - 1.0 / (self.zoom * self.ar),
                self.pan[1] + 1.0 / (self.zoom * self.flat_earth))

    def on_panzoom(self, data=None, finished=True):
        if not self.initialized:
            return False

        # Don't pan further than the poles in y-direction
        self.pan[0] = min(max(self.pan[0], -90.0 + 1.0 /
            (self.zoom * self.ar)), 90.0 - 1.0 / (self.zoom * self.ar))

        # Update flat-earth factor and possibly zoom in case of very wide windows (> 2:1)
        self.flat_earth = np.cos(np.deg2rad(self.pan[0]))

        # Limit zoom extents in x-direction to [-180:180], and in y-direction to [-90:90]
        self.zoom = max(self.zoom, 1.0 / min(90.0 *
                                             self.ar, 180.0 * self.flat_earth))

        # Check for necessity wrap-around in x-direction
        self.wraplon = -999.9
        self.wrapdir = 0
        if self.pan[1] + 1.0 / (self.zoom * self.flat_earth) < -180.0:
            # The left edge of the map has passed the right edge of the screen: we can just change the pan position
            self.pan[1] += 360.0
        elif self.pan[1] - 1.0 / (self.zoom * self.flat_earth) < -180.0:
            # The left edge of the map has passed the left edge of the screen: we need to wrap around to the left
            self.wraplon = float(
                np.ceil(360.0 + self.pan[1] - 1.0 / (self.zoom * self.flat_earth)))
            self.wrapdir = -1
        elif self.pan[1] - 1.0 / (self.zoom * self.flat_earth) > 180.0:
            # The right edge of the map has passed the left edge of the screen: we can just change the pan position
            self.pan[1] -= 360.0
        elif self.pan[1] + 1.0 / (self.zoom * self.flat_earth) > 180.0:
            # The right edge of the map has passed the right edge of the screen: we need to wrap around to the right
            self.wraplon = float(
                np.floor(-360.0 + self.pan[1] + 1.0 / (self.zoom * self.flat_earth)))
            self.wrapdir = 1

        # Use new centre lat/lon to update reference position
        bs.ref.lat = self.pan[0]
        bs.ref.lon = self.pan[1]
        bs.ref.area = areafilter.Box('refarea', self.viewportlatlon())

        # update pan and zoom on GPU for all shaders
        self.shaderset.set_wrap(self.wraplon, self.wrapdir)
        self.shaderset.set_pan_and_zoom(self.pan[0], self.pan[1], self.zoom)

    def setpanzoom(self, pan=None, zoom=None, origin=None, absolute=True, finished=True):
        # Absolute or relative pan operation
        if pan is not None:
            self.pan = pan if absolute else [i + j for i, j in zip(self.pan, pan)]
        if zoom is not None:
            if absolute:
                # Limit zoom extents in x-direction to [-180:180], and in y-direction to [-90:90]
                self.zoom = max(
                    zoom, 1.0 / min(90.0 * self.ar, 180.0 * self.flat_earth))
            else:
                prevzoom = self.zoom
                glx, gly = self.pixelCoordsToGLxy(
                    *origin) if origin else (0, 0)
                self.zoom *= zoom

                # Limit zoom extents in x-direction to [-180:180], and in y-direction to [-90:90]
                self.zoom = max(self.zoom, 1.0 / min(90.0 *
                                                        self.ar, 180.0 * self.flat_earth))

                # Correct pan so that zoom actions are around the mouse position, not around 0, 0
                # glxy / zoom1 - pan1 = glxy / zoom2 - pan2
                # pan2 = pan1 + glxy (1/zoom2 - 1/zoom1)
                self.pan[1] = self.pan[1] - glx * \
                    (1.0 / self.zoom - 1.0 / prevzoom) / self.flat_earth
                self.pan[0] = self.pan[0] - gly * \
                    (1.0 / self.zoom - 1.0 / prevzoom) / self.ar
        ss.get().panzoom.zoom = self.zoom #  temp TODO
        ctx.topic = 'PANZOOM'
        self.panzoom_event.emit(ss.get().panzoom, finished)
        ctx.topic = None
        return True

    def event(self, event):
        ''' Event handling for input events. '''
        if event.type() == QEvent.Type.Wheel:
            # For mice we zoom with control/command and the scrolwheel
            if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                origin = (event.position().x(), event.position().y())
                zoom = 1.0
                try:
                    if event.pixelDelta():
                        # High resolution scroll
                        zoom *= (1.0 + 0.01 * event.pixelDelta().y())
                    else:
                        # Low resolution scroll
                        zoom *= (1.0 + 0.001 * event.angleDelta().y())
                except AttributeError:
                    zoom *= (1.0 + 0.001 * event.delta())
                self.panzoomchanged = True
                return self.setpanzoom(zoom=zoom, origin=origin, absolute=False, finished=False)

            # For touchpad scroll (2D) is used for panning
            else:
                try:
                    dlat = 0.01 * event.pixelDelta().y() / (self.zoom * self.ar)
                    dlon = -0.01 * event.pixelDelta().x() / (self.zoom * self.flat_earth)
                    self.panzoomchanged = True
                    return self.setpanzoom(pan=[dlat, dlon], absolute=False, finished=False)
                except AttributeError:
                    pass

        # For touchpad, pinch gesture is used for zoom
        elif event.type() == QEvent.Type.Gesture:
            pan = zoom = None
            dlat = dlon = 0.0
            for g in event.gestures():
                if g.gestureType() == Qt.GestureType.PinchGesture:
                    event.accept(g)
                    zoom = g.scaleFactor() * (zoom or 1.0)

                elif g.gestureType() == Qt.GestureType.PanGesture:
                    event.accept(g)
                    if abs(g.delta().y() + g.delta().x()) > 1e-1:
                        dlat += 0.005 * g.delta().y() / (self.zoom * self.ar)
                        dlon -= 0.005 * g.delta().x() / (self.zoom * self.flat_earth)
                        pan = (dlat, dlon)
            if pan is not None or zoom is not None:
                self.panzoomchanged = True
                return self.setpanzoom(pan, zoom, origin=self.mousepos, absolute=False, finished=False)

        elif event.type() == QEvent.Type.MouseButtonPress and event.button() & Qt.MouseButton.LeftButton:
            self.mousedragged = False
            # For mice we pan with control/command and mouse movement.
            # Mouse button press marks the beginning of a pan
            self.prevmousepos = (event.pos().x(), event.pos().y())

        elif event.type() == QEvent.Type.MouseButtonRelease and \
                event.button() & Qt.MouseButton.LeftButton and not self.mousedragged:
            lat, lon = self.pixelCoordsToLatLon(event.pos().x(), event.pos().y())
            self.radarclick_event.emit(lat, lon)

        elif event.type() == QEvent.Type.MouseMove:
            self.mousedragged = True
            self.mousepos = (event.pos().x(), event.pos().y())
            if event.buttons() & Qt.MouseButton.LeftButton:
                dlat = 0.003 * \
                    (event.pos().y() - self.prevmousepos[1]) / (self.zoom * self.ar)
                dlon = 0.003 * \
                    (self.prevmousepos[0] - event.pos().x()) / \
                    (self.zoom * self.flat_earth)
                self.prevmousepos = (event.pos().x(), event.pos().y())
                self.panzoomchanged = True

                return self.setpanzoom(pan=(dlat, dlon), absolute=False, finished=False)

        elif event.type() == QEvent.Type.TouchBegin:
            # Accept touch start to enable reception of follow-on touch update and touch end events
            event.accept()

        # Update pan/zoom to simulation thread only when the pan/zoom gesture is finished
        elif (event.type() == QEvent.Type.MouseButtonRelease or
              event.type() == QEvent.Type.TouchEnd) and self.panzoomchanged:
            self.panzoomchanged = False
            bs.net.send(b'PANZOOM', dict(pan=(self.pan[0], self.pan[1]),
                                         zoom=self.zoom, ar=self.ar, absolute=True))
            self.panzoom_event.emit(ss.get().panzoom, True)
        elif int(event.type()) == 216:
            # 216 is screen change event, but doesn't exist (yet) in pyqt as enum
            self.pxratio = self.devicePixelRatio()
            self.shaderset.set_pixel_ratio(self.pxratio)
            return super().event(event)
        else:
            return super().event(event)
        
        # If we get here, the event was a mouse/trackpad event. Emit it to interested children
        self.mouse_event.emit(event)

        # For all other events call base class event handling
        return True




import numpy as np

from PyQt6.QtWidgets import QListView, QStyledItemDelegate, QApplication as app
from PyQt6.QtGui import QColor, QPalette
from PyQt6.QtCore import QSize, Qt, QRect, QAbstractListModel, QModelIndex, Qt, QVariant, QTimer

from bluesky.core import Base
from bluesky.network.subscriber import subscriber
from bluesky.network.sharedstate import ActData
from bluesky.tools import aero


class TrafficModel(QAbstractListModel, Base):
    id: ActData[list] = ActData(group='acdata')
    alt: ActData[np.ndarray] = ActData(0, group='acdata')
    trk: ActData[np.ndarray] = ActData(0, group='acdata')
    cas: ActData[np.ndarray] = ActData(0, group='acdata')

    def __init__(self) -> None:
        super().__init__()
        self.naircraft = 0

        timer = QTimer(self)
        timer.timeout.connect(self.notifyUpdate)
        timer.start(1000)

    def notifyUpdate(self):
        if self.naircraft:
            self.dataChanged.emit(self.index(0, 0), self.index(self.rowCount(), 0))

    def rowCount(self, parent=QModelIndex()):
        return len(self.id)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if index.isValid():
            # if role == Qt.ItemDataRole.DisplayRole:
            idx = index.row()
            # Qt is multi-threaded, so data model size can change during view update
            if idx >= len(self.id):
                return QVariant()
            return QVariant((
                self.id[idx],
                f'FL{self.alt[idx] / aero.ft / 100:1.0f}',
                f'{self.trk[idx]:1.0f}',
                f'{self.cas[idx] / aero.kts:1.0f}'
            ))

    @subscriber(topic='ACDATA', actonly=True)
    def on_data_update(self, data):
        if len(data.id) == 0:
            self.beginResetModel()
            self.endResetModel()
        elif len(data.id) > self.naircraft:
            self.beginInsertRows(QModelIndex(), self.naircraft, len(data.id) - 1)
            self.endInsertRows()
        elif len(data.id) < self.naircraft:
            diff = self.naircraft - len(data.id)
            self.beginRemoveRows(QModelIndex(), self.naircraft - diff - 1, self.naircraft - 1)
            self.endRemoveRows()
        self.naircraft = len(data.id)


class TrafficList(QListView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setModel(TrafficModel())
        self.setBackgroundRole(QPalette.ColorRole.NoRole)
        self.setAutoFillBackground(True)
        self.setStyleSheet('background-color: transparent')
        self.setItemDelegate(TrafficItem())
    

class TrafficItem(QStyledItemDelegate):
    """ AreaItem is used to render area items in the area list """
    def paint(self, painter, option, index):
        # Determine dark mode
        p = app.instance().style().standardPalette()
        isdark = (p.color(p.ColorRole.Window).value() < p.color(p.ColorRole.WindowText).value())

        if isdark:
            bgcolor = p.color(p.ColorRole.Window).lighter(120)
            bghover = p.color(p.ColorRole.Window).lighter(140)
            txtcolor = Qt.GlobalColor.white
        else:
            bgcolor = QColor(250, 250, 250, 255)
            bghover = Qt.GlobalColor.lightGray
            txtcolor = Qt.GlobalColor.black

        # Get polygon name and data
        acid, alt, trk, cas = index.data()

        r = QRect(option.rect)
        r.setLeft(r.left() + 15)

        font = painter.font()
        font.setBold(True)
        painter.setFont(font)
        painter.drawText(r, Qt.AlignmentFlag.AlignLeft, acid)
        bb = painter.boundingRect(option.rect, Qt.AlignmentFlag.AlignLeft, acid)
        font.setBold(False)
        painter.setFont(font)
        r.setLeft(r.left() + bb.width())
        painter.drawText(r, Qt.AlignmentFlag.AlignLeft, ' '.join((alt, trk, cas)))

    def sizeHint(self, option, index):
        """ Returns the size needed to display the item in a QSize object. """
        # index.data()
        return QSize(200, 25)
    


''' Text parsing utilities for fixed-width column data'''
# Use parser = FixedWidthParser(specformat)
#
# List of formats strings, one per string per line, each string contains field formats:
#
#   First argument per line is a discriminator, it searches first line with this key
#
#      3x = skip 3 places
#      6S = string of 6 chars
#     10F = float of 10 chars
#      5I = integer of 5 chars
#
#     'CD, 3X, 6S, 9X, 1I, 12X, 9S, 17X, 1S',  # means find first line starting with CD, then skip 3 columns,
#     'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
#     'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
#     'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
#
# First create a parser for a file format, then you can read different files with this parser,
# by calling the parse(file) method

import re

# Define regular expressions for finding:
# The types in a dataformat string
re_types = re.compile('[.*[fis]]*', re.IGNORECASE)
# The ranges that need to be skipped (indicated with [0-9]X)
re_skip  = re.compile('[\s,]*(\d+)x[\s,]*', re.IGNORECASE)
# The ranges that contain data (indicated with [0-9][fis])
re_data  = re.compile('[\s,]*(\d+)[fis][\s,]*', re.IGNORECASE)


class FixedWidthParser:
    ''' Use a FixedWidthParser to parse fixed-width column-based text files.
        In BlueSky, this is currently used to parse BADA data files.

        The column format should be specified in a table passed to the constructor
        of FixedWidthParser. '''

    # Data elements can be either floats, ints, or strings.
    types = {'f': float, 'i': int, 's': str}

    def __init__(self, specformat):
        self.dformat  = []
        # If the provided dataformat only has one line assume that all lines
        # have the same column format
        self.repeat   = len(specformat) == 1
        # Split the specified dataformat in a table of types to convert the
        # parsed text with, and a list of regular expression objects that will
        # be used to extract the fixed-with data columns.
        for line in specformat:
            line_re   = re.compile(re_data.sub(r'(.{\1})', re_skip.sub(r'.{\1}', line)))
            linetypes = [self.types[t.lower()] for t in re_types.findall(line)]
            self.dformat.append((line_re, linetypes))

    def parse(self, fname):
        line_re, dtypes = self.dformat[0]
        data            = []
        with open(fname) as f:
            for lineno, line in enumerate(f):
                match = line_re.match(line)
                if match:
                    try:
                        dline = [t(s.strip()) for t, s in zip(dtypes, match.groups())]
                    except:
                        raise ParseError(fname, lineno + 1)
                    data.append(dline)
                    if not self.repeat:
                        # If we've extracted all datalines we can close the file
                        if len(data) == len(self.dformat):
                            break
                        # otherwise select the next dataline parser
                        line_re, dtypes = self.dformat[len(data)]
        return data

class ParseError(Exception):
    def __init__(self, fname, lineno):
        super().__init__()
        self.fname = fname
        self.lineno = lineno



''' Sim-side implementation of graphical data plotter in BlueSky.'''
from collections import defaultdict
import bluesky as bs
from bluesky.core import varexplorer as ve


# Globals
# The list of plots
plots = list()


def plot(*args, **params):
    ''' Select a set of variables to plot.
        Arguments: varx, vary, dt, color, fig. '''
    if args:
        try:
            newplot = Plot(*args, **params)
            plots.append(newplot)
        except IndexError as e:
            return False, e.args[0]
    bs.net.send(b'PLOT', dict(show=True), (bs.stack.sender() or b'*'))
    return True


def legend(legend, fig=None):
    ''' Set a legend for a figure. '''
    try:
        # Get the plot with the corresponding figure number
        p = plots[-1] if fig is None else next(
            plot for plot in plots if plot.fig == str(fig))

        data = {p.fig: dict(legend=legend)}
        bs.net.send(b'PLOT', data, to_group=p.to_group)
        return True
    except (IndexError, StopIteration) as e:
        return False, e.args[0]

def reset():
    ''' Remove plots when simulation is reset. '''
    # Notify clients of removal of plots
    notify_ids = {p.to_group for p in plots}
    for to_group in notify_ids:
        bs.net.send(b'PLOT', dict(reset=True), to_group)
    plots.clear()


def update():
    ''' Periodic update function for the plotter. '''
    streamdata = defaultdict(dict)
    for p in plots:
        if p.tnext <= bs.sim.simt:
            p.tnext += p.dt
            streamdata[p.to_group][p.fig] = dict(x=p.x.get(), y=p.y.get())

    for to_group, data in streamdata.items():
        bs.net.send(b'PLOT', data, to_group)


class Plot:
    ''' A plot object.
        Each plot object is used to manage the plot of one variable
        on the sim side.'''

    maxfig = 0

    def __init__(self, varx='', vary='', dt=1.0, fig=None, **params):
        self.x = ve.findvar(varx if vary else 'simt')
        self.y = ve.findvar(vary or varx)
        self.dt = dt
        self.tnext = bs.sim.simt
        self.params = params
        if not fig:
            fig = Plot.maxfig
            Plot.maxfig += 1
        elif fig > Plot.maxfig:
            Plot.maxfig = fig

        self.fig = str(fig)

        self.to_group = (bs.stack.sender() or b'*')

        if None in (self.x, self.y):
            raise IndexError('Variable {} not found'.format(varx if self.x is None else (vary or varx)))

        # if not self.x.is_num() or not self.y.is_num():
        #     raise IndexError('Variable {} not numeric'.format(varx if not self.x.is_num() else (vary or varx)))
        bs.net.send(b'PLOT', {self.fig: params}, self.to_group)

    def send(self):
        bs.net.send(b'PLOT', {self.fig : dict(x=self.x.get(), y=self.y.get())}, self.to_group)



""" Create aircraft trails on the radar display."""
from math import *
import numpy as np
import bluesky as bs
from bluesky import settings
from bluesky.core import TrafficArrays


class Trails(TrafficArrays):
    """
    Traffic trails class definition    : Data for trails

    Methods:
        Trails()            :  constructor

    Members: see create

    Created by  : Jacco M. Hoekstra
    """

    def __init__(self,dttrail=10.):
        super().__init__()
        self.active = False  # Wether or not to show trails
        self.dt = dttrail    # Resolution of trail pieces in time
        self.pygame = (bs.gui == 'pygame') # Trails are different for pygame
        self.tcol0 = 60.  # After how many seconds old colour

        # This list contains some standard colors
        self.colorList = {'BLUE': np.array([0, 0, 255]),
                          'CYAN': np.array([0,255,255]),
                          'RED' : np.array([255, 0, 0]),
                          'YELLOW': np.array([255, 255, 0])}

        # Set default color to Blue
        self.defcolor = self.colorList['CYAN']

        # Foreground data on line pieces
        self.lat0 = np.array([])
        self.lon0 = np.array([])
        self.lat1 = np.array([])
        self.lon1 = np.array([])
        self.time = np.array([])
        self.col  = []
        self.fcol = np.array([])

        # background copy of data
        self.bglat0 = np.array([])
        self.bglon0 = np.array([])
        self.bglat1 = np.array([])
        self.bglon1 = np.array([])
        self.bgtime = np.array([])
        self.bgcol = []

        with self.settrafarrays():
            self.accolor = []
            self.lastlat = np.array([])
            self.lastlon = np.array([])
            self.lasttim = np.array([])

        self.clearnew()

        return

    def create(self,n=1):
        super().create(n)

        self.accolor[-1] = self.defcolor
        self.lastlat[-1] = bs.traf.lat[-1]
        self.lastlon[-1] = bs.traf.lon[-1]

    def update(self):
        self.acid    = bs.traf.id
        if not self.active:
            self.lastlat = bs.traf.lat
            self.lastlon = bs.traf.lon
            self.lasttim[:] = bs.sim.simt
            return
        """Add linepieces for trails based on traffic data"""

        # Use temporary list/array for fast append
        lstlat0 = []
        lstlon0 = []
        lstlat1 = []
        lstlon1 = []
        lsttime = []

        # Check for update
        delta = bs.sim.simt - self.lasttim
        idxs = np.where(delta > self.dt)[0]

        # Add all a/c which need the update
        # if len(idxs)>0:
        #     print "len(idxs)=",len(idxs)

        for i in idxs:
            # Add to lists
            lstlat0.append(self.lastlat[i])
            lstlon0.append(self.lastlon[i])
            lstlat1.append(bs.traf.lat[i])
            lstlon1.append(bs.traf.lon[i])
            lsttime.append(bs.sim.simt)

            if isinstance(self.col, np.ndarray):
                # print type(trailcol[i])
                # print trailcol[i]
                # print "col type: ",type(self.col)
                self.col = self.col.tolist()

            type(self.col)
            self.col.append(self.accolor[i])

            # Update aircraft record
            self.lastlat[i] = bs.traf.lat[i]
            self.lastlon[i] = bs.traf.lon[i]
            self.lasttim[i] = bs.sim.simt

        # When a/c is no longer part of trail semgment,
        # it is no longer a/c data => move to the GUI buffer (send or draw)
        if self.pygame:
            # Pygame: send to drawing buffer
            self.lat0 = np.concatenate((self.lat0, np.array(lstlat0)))
            self.lon0 = np.concatenate((self.lon0, np.array(lstlon0)))
            self.lat1 = np.concatenate((self.lat1, np.array(lstlat1)))
            self.lon1 = np.concatenate((self.lon1, np.array(lstlon1)))
            self.time = np.concatenate((self.time, np.array(lsttime)))
        else:
            # QtGL: add to send buffer
            self.newlat0.extend(lstlat0)
            self.newlon0.extend(lstlon0)
            self.newlat1.extend(lstlat1)
            self.newlon1.extend(lstlon1)
        # Update colours
        self.fcol = (1. - np.minimum(self.tcol0, np.abs(bs.sim.simt - self.time)) / self.tcol0)

        return

    def buffer(self):
        """Buffer trails: Move current stack to background """

        self.bglat0 = np.append(self.bglat0, self.lat0)
        self.bglon0 = np.append(self.bglon0, self.lon0)
        self.bglat1 = np.append(self.bglat1, self.lat1)
        self.bglon1 = np.append(self.bglon1, self.lon1)
        self.bgtime = np.append(self.bgtime, self.time)

        # No color saved: Background: always 'old color' self.col0
        if isinstance(self.bgcol, np.ndarray):
            self.bgcol = self.bgcol.tolist()
        if isinstance(self.col, np.ndarray):
            self.col = self.col.tolist()

        self.bgcol = self.bgcol + self.col
        self.bgacid = self.bgacid + self.acid

        self.clearfg()  # Clear foreground trails
        return

    def clearnew(self):
        # Clear new lines pipeline used for QtGL
        self.newlat0 = []
        self.newlon0 = []
        self.newlat1 = []
        self.newlon1 = []


    def clearfg(self):  # Foreground
        """Clear trails foreground"""
        self.lat0 = np.array([])
        self.lon0 = np.array([])
        self.lat1 = np.array([])
        self.lon1 = np.array([])
        self.time = np.array([])
        self.col = np.array([])
        return

    def clearbg(self):  # Background
        """Clear trails background"""
        self.bglat0 = np.array([])
        self.bglon0 = np.array([])
        self.bglat1 = np.array([])
        self.bglon1 = np.array([])
        self.bgtime = np.array([])
        self.bgacid = []
        return

    def clear(self):
        """Clear all data, Foreground and background"""
        self.lastlon = np.array([])
        self.lastlat = np.array([])
        self.clearfg()
        self.clearbg()
        self.clearnew()
        return

    def setTrails(self, *args):
        """ Set trails on/off, or change trail color of aircraft """
        if len(args)==0:
            msg = "TRAIL ON/OFF, [dt] / TRAIL acid color\n"

            if self.active:
                msg = msg + "TRAILS ARE ON"
            else:
                msg = msg + "TRAILS ARE OFF"

            return True,msg

        # Switch on/off
        elif type(args[0]) == bool:
            # Set trails on/off
            self.active = args[0]
            if len(args) > 1:
                self.dt = args[1]
            if not self.active:
                self.clear()

        # Change color per acid (pygame only)
        else:

            # Change trail color
            if len(args) < 2 or args[1] not in ["BLUE", "RED", "YELLOW"]:
                return False, "Set aircraft trail color with: TRAIL acid BLUE/RED/YELLOW"
            self.changeTrailColor(args[1], args[0])

        return True

    def changeTrailColor(self, color, idx):
        """Change color of aircraft trail"""
        self.accolor[idx] = self.colorList[color]
        return

    def reset(self):
        # This ensures that the traffic arrays (which size is dynamic)
        # are all reset as well, so all lat,lon,sdp etc but also objects adsb
        super().reset()
        self.clear()
        self.active = False



""" BlueSky plugin template. The text you put here will be visible
    in BlueSky as the description of your plugin. """
from random import randint
import numpy as np
# Import the global bluesky objects. Uncomment the ones you need
from bluesky import core, stack, traf  #, settings, navdb, sim, scr, tools

### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():
    ''' Plugin initialisation function. '''
    # Instantiate our example entity
    example = Example()

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'EXAMPLE',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim',
        }

    # init_plugin() should always return a configuration dict.
    return config


### Entities in BlueSky are objects that are created only once (called singleton)
### which implement some traffic or other simulation functionality.
### To define an entity that ADDS functionality to BlueSky, create a class that
### inherits from bluesky.core.Entity.
### To replace existing functionality in BlueSky, inherit from the class that
### provides the original implementation (see for example the asas/eby plugin).
class Example(core.Entity):
    ''' Example new entity object for BlueSky. '''
    def __init__(self):
        super().__init__()
        # All classes deriving from Entity can register lists and numpy arrays
        # that hold per-aircraft data. This way, their size is automatically
        # updated when aircraft are created or deleted in the simulation.
        with self.settrafarrays():
            self.npassengers = np.array([])

    def create(self, n=1):
        ''' This function gets called automatically when new aircraft are created. '''
        # Don't forget to call the base class create when you reimplement this function!
        super().create(n)
        # After base creation we can change the values in our own states for the new aircraft
        self.npassengers[-n:] = [randint(0, 150) for _ in range(n)]

    # Functions that need to be called periodically can be indicated to BlueSky
    # with the timed_function decorator
    @core.timed_function(name='example', dt=5)
    def update(self):
        ''' Periodic update function for our example entity. '''
        stack.stack('ECHO Example update: creating a random aircraft')
        stack.stack('MCRE 1')

    # You can create new stack commands with the stack.command decorator.
    # By default, the stack command name is set to the function name.
    # The default argument type is a case-sensitive word. You can indicate different
    # types using argument annotations. This is done in the below function:
    # - The acid argument is a BlueSky-specific argument with type 'acid'.
    #       This converts callsign to the corresponding index in the traffic arrays.
    # - The count argument is a regular int.
    @stack.command
    def passengers(self, acid: 'acid', count: int = -1):
        ''' Set the number of passengers on aircraft 'acid' to 'count'. '''
        if count < 0:
            return True, f'Aircraft {traf.id[acid]} currently has {self.npassengers[acid]} passengers on board.'

        self.npassengers[acid] = count
        return True, f'The number of passengers on board {traf.id[acid]} is set to {count}.'



''' BlueSky command-line argument parsing. '''
import argparse

from bluesky.network.common import hex2bin

def parse():
    parser = argparse.ArgumentParser(prog="BlueSky", description="   *****   BlueSky Open ATM simulator *****")

    def setmodegui(mode, gui=None, discoverable=None, detached=None):
        class ModeGuiAction(argparse.Action):
            def __call__(self, parser, namespace, values=None, option_string=''):
                namespace.mode = mode
                namespace.gui = gui
                namespace.hostname = values or None
                if discoverable is not None:
                    namespace.discoverable = discoverable
                if detached is not None:
                    namespace.detached = detached

        return ModeGuiAction

    class GroupidAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            namespace.group_id = hex2bin(values)

    # Add all possible arguments to bluesky here
    mode = parser.add_mutually_exclusive_group()
    mode.set_defaults(mode="server", gui="qtgl", hostname=None, detached=False)
    mode.add_argument("--headless", dest="mode", action=setmodegui("server", discoverable=True), nargs=0, help="Start simulation server only, without GUI.")
    mode.add_argument("--client", dest="hostname", action=setmodegui("client", "qtgl"),
                    nargs="?", default=None, help="Start QtGL graphical client, which can connect to an already running server. When no hostname is passed, a discovery dialog is shown to let the user select a BlueSky server")
    mode.add_argument("--console", dest="hostname", action=setmodegui("client", "console"),
                    nargs="?", default=None, help="Start console client, which can connect to an already running server. When no hostname is passed, a discovery dialog is shown to let the user select a BlueSky server")
    mode.add_argument("--sim", dest="mode", action=setmodegui("sim"),
                    nargs=0, help="Start only one simulation node.")
    mode.add_argument("--detached", dest="mode", action=setmodegui("sim", detached=True),
                    nargs=0, help="Start only one simulation node, without networking.")


    parser.add_argument("--configfile", dest="configfile",
                        help="Load an alternative configuration file.")

    scnparse = parser.add_mutually_exclusive_group()
    scnparse.add_argument(dest="scenfile", nargs="?", default=argparse.SUPPRESS,
                        help="Load scenario file on startup.")
    scnparse.add_argument("--scenfile", dest="scenfile",
                        help="Load scenario file on startup.")

    parser.add_argument("--discoverable", dest="discoverable", action="store_const", const=True,
                        default=False, help="Make simulation server discoverable. (Default in headless mode).")
    
    parser.add_argument("--workdir", dest="workdir",
                        help="Set BlueSky working directory (if other than cwd or ~/bluesky).")

    parser.add_argument("--groupid", dest="group_id", action=GroupidAction,
                        help="Explicitly set a group part, or the entire connection id instead of randomly generating it.")

    cmdargs = parser.parse_args()

    return vars(cmdargs)



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPen, QBrush, QColor, QFont
from PyQt6.QtWidgets import QGraphicsView, QGraphicsScene, QGraphicsItemGroup


class AMANDisplay(QGraphicsView):
    def __init__(self):
        super().__init__()
        self.setGeometry(0, 0, 500, 600)
        self.setStyleSheet('background-color:#233370')
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.gfxscene = QGraphicsScene(0, 0, 500, 600)

        # Timeline boundaries
        pen = QPen(QColor('white'))
        brush = QBrush(QColor('#233370'))
        self.gfxscene.addLine(220, 0, 220, 600, pen)
        self.gfxscene.addLine(280, 0, 280, 600, pen)
        self.gfxscene.addLine(0, 30, 500, 30, pen)

        timelinebox = self.gfxscene.addRect(220, 30, 60, 540, pen, brush)
        timelinebox.setFlag(timelinebox.ItemClipsChildrenToShape, True)

        # Timeline scale
        self.timeline  = QGraphicsItemGroup()
        self.timeline.setParentItem(timelinebox)
        self.timeticks = []
        for i in range(40):
            y = 15 * i
            w = 6
            if i % 5 == 0:
                w = 10
                self.timeticks.append(self.gfxscene.addText('%02d' % (40 - i), QFont('Courier', 10)))
                self.timeticks[-1].setPos(240, y - 10)
                self.timeticks[-1].setDefaultTextColor(QColor('white'))
                self.timeline.addToGroup(self.timeticks[-1])
            self.timeline.addToGroup(self.gfxscene.addLine(220, y, 220 + w, y, pen))
            self.timeline.addToGroup(self.gfxscene.addLine(280 - w, y, 280, y, pen))

        self.lrwy = self.gfxscene.addText('18R', QFont('Arial', 20, 50))
        self.lrwy.setPos(1, 1)
        self.lrwy.setDefaultTextColor(QColor('white'))
        # Finalize
        self.setScene(self.gfxscene)
        self.show()

    def update(self, simt, data):
        # data has: ids, iafs, eats, etas, delays, rwys, spdratios

        # First delete old labels
        for key, in self.aircraft.keys:
            if key not in data.ids:
                # del label
                del self.aircraft[key]

        for i in len(data.ids):
            if data.ids[i] not in self.aircraft:
                # self.aircraft[data.ids[i]] = QLabel()
                pass

            # Generate the new label text and position
            newtext = '<font color=Red>bla</font>'  # veranderen
            lbl = self.aircraft[data.ids[i]]
            lbl.setText(newtext)
            lbl.move(posx, posy)  # move in pixels



''' Network context namespace. 

    Gives access to original message context for connected functions.
'''
from typing import Any, Optional
from bluesky.network.common import ActionType


# For shared state messages: the action type
action: ActionType = ActionType.NoAction

# For shared state messages: the action content
action_content: Optional[Any] = None

# The node id of the sender of the current message
sender_id = b''

# The topic of the current message
topic = ''

# The raw message
msg = b''



''' BlueSky simulation control object. '''
import time
import datetime
import signal
import numpy as np
from random import seed

# Local imports
import bluesky as bs
from bluesky.core.base import Base
import bluesky.core as core
from bluesky.core import plugin, simtime
from bluesky.core.signal import subscriber
from bluesky.network.publisher import state_publisher
from bluesky.core.walltime import Timer
from bluesky.core.timedfunction import hooks
from bluesky.stack import simstack, recorder
from bluesky.tools import datalog, areafilter, plotter


# Minimum sleep interval
MINSLEEP = 1e-3

# Register settings defaults
bs.settings.set_variable_defaults(simdt=0.05)

class Simulation(Base):
    ''' The simulation object. '''
    def __init__(self):
        super().__init__()
        self.state = bs.INIT
        self.prevstate = None

        # System time [seconds]
        self.syst = -1.0

        # Benchmark time and timespan [seconds]
        self.bencht = 0.0
        self.benchdt = -1.0

        # Simulation time [seconds]
        self.simt = 0.0

        # Simulation timestep [seconds]
        self.simdt = bs.settings.simdt

        # Simulation timestep multiplier: run sim at n x speed
        self.dtmult = 1.0

        # Simulated UTC clock time
        self.utc = datetime.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

        # Flag indicating running at fixed rate or fast time
        self.ffmode = False
        self.ffstop = None

        # Flag indicating whether timestep can be varied to ensure realtime op
        self.rtmode = False

        self.running = True

        # Keep track of known clients
        self.clients = set()

        # Connect to system ABORT/INTERRUPT signal
        signal.signal(signal.SIGINT, lambda *args: self.quit())
        signal.signal(signal.SIGTERM, lambda *args: self.quit())

    def run(self):
        ''' Start the main loop of this simulation. '''
        while self.running:
            # Process timers
            Timer.update_timers()
            # Update network connections
            bs.net.update()
            # Perform a simulation step
            self.update()

            # Update screen logic
            bs.scr.update()

        # Close up properly on exit
        bs.net.close()
        datalog.reset()

        # Close savefile which may be open for recording
        recorder.saveclose()  # Close reording file if it is on

    def step(self, dt_increment=0):
        ''' Perform one simulation timestep.
        
            Call this function instead of update if you don't want to run with a fixed
            real-time rate.
        '''
        if self.state == bs.INIT:
            # Simulation starts as soon as there is traffic, or pending commands
            if bs.traf.ntraf > 0 or len(bs.stack.get_scendata()[0]) > 0:
                self.op()

        # Always update stack
        simstack.process()

        if self.state == bs.OP:
            # Plot/log the current timestep, and call preupdate functions
            plotter.update()
            datalog.update()
            hooks.preupdate.trigger()

            # Determine interval towards next timestep                
            self.simt, self.simdt = simtime.step(dt_increment)

            # Update UTC time
            self.utc += datetime.timedelta(seconds=self.simdt)

            # Update traffic and other update functions for the next timestep
            bs.traf.update()
            hooks.update.trigger()

        else:
            hooks.hold.trigger()

    def update(self):
        ''' Perform a simulation update. 
            This involves performing a simulation step, and when running in real-time mode
            (or a multiple thereof), sleeping an appropriate time. '''
        if self.state == bs.INIT:
            if self.syst < 0.0:
                self.syst = time.time()

            if self.benchdt > 0.0:
                self.fastforward(self.benchdt)
                self.bencht = time.time()

        # When running at a fixed rate, or when in hold/init,
        # increment system time with sysdt and calculate remainder to sleep.
        remainder = self.syst - time.time()
        if (not self.ffmode or self.state != bs.OP) and remainder > MINSLEEP:
            time.sleep(remainder)

        # Perform one simulation timestep
        if remainder < 0.0 and self.rtmode:
            # Allow a variable timestep when we are running realtime
            self.step(-remainder)
        else:
            # Don't accumulate delay when we aren't running realtime
            if remainder < 0:
                self.syst -= remainder
            self.step()

        # Always update syst
        self.syst += self.simdt / self.dtmult

        # Stop fast-time/benchmark if enabled and set interval has passed
        if self.ffstop is not None and self.simt >= self.ffstop:
            if self.benchdt > 0.0:
                bs.scr.echo('Benchmark complete: %d samples in %.3f seconds.' %
                            (bs.scr.samplecount, time.time() - self.bencht))
                self.benchdt = -1.0
                self.hold()
            else:
                self.op()


        # Inform main of our state change
        if self.state != self.prevstate:
            bs.net.send(b'STATECHANGE', self.state, bs.net.server_id)
            self.prevstate = self.state

    def quit(self):
        ''' Quit simulation.
            This function is called when a QUIT signal is received from
            the server, or when quit is called. '''
        self.state = bs.END
        self.running = False

    def op(self):
        ''' Set simulation state to OPERATE. '''
        self.syst = time.time() + self.simdt
        self.ffmode = False
        self.ffstop = None
        self.state = bs.OP
        self.set_dtmult(1.0)

    def hold(self):
        ''' Set simulation state to HOLD. '''
        self.syst = time.time() + self.simdt / self.dtmult
        self.state = bs.HOLD
        self.ffmode = False
        self.ffstop = None


    def reset(self):
        ''' Reset all simulation objects. '''
        self.state = bs.INIT
        self.syst = -1.0
        self.simt = 0.0
        self.simdt = bs.settings.simdt
        simtime.reset()
        self.utc = datetime.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
        self.ffmode = False
        self.set_dtmult(1.0)
        hooks.reset.trigger()
        core.reset()
        bs.ref.reset()
        bs.navdb.reset()
        bs.traf.reset()
        simstack.reset()
        datalog.reset()
        areafilter.reset()
        bs.scr.reset()
        plotter.reset()

        # Communicate that this simulation has reset
        bs.net.send(b'RESET')

    def set_dtmult(self, mult):
        ''' Set simulation speed multiplier. '''
        self.dtmult = mult

    def realtime(self, flag=None):
        if flag is not None:
            self.rtmode = flag

        return True, 'Realtime mode is o' + ('n' if self.rtmode else 'ff')

    def fastforward(self, nsec=None):
        ''' Run in fast-time (for nsec seconds if specified). '''
        self.state = bs.OP
        self.ffmode = True
        self.ffstop = (self.simt + nsec) if nsec else None

    def benchmark(self, fname='IC', dt=300.0):
        ''' Run a simulation benchmark.
            Use scenario given by fname.
            Run for <dt> seconds. '''
        simstack.ic(fname)
        self.bencht  = 0.0  # Start time will be set at next sim cycle
        self.benchdt = dt

    def batch(self, fname):
        ''' Run a batch of scenarios. '''
        # The contents of the scenario file are meant as a batch list:
        # send to server and clear stack
        self.reset()
        try:
            scentime, scencmd = zip(*[tc for tc in simstack.readscn(fname)])
            bs.net.send(b'BATCH', (scentime, scencmd), bs.net.server_id)
        except FileNotFoundError:
            return False, f'BATCH: File not found: {fname}'

        return True

    @subscriber(topic='BATCH')
    def start_batch_scenario(self, data):
        ''' Start a scenario coming from the server batch. '''
        # We are in a batch simulation, and received an entire scenario. Assign it to the stack.
        self.reset()
        bs.stack.set_scendata(data['scentime'], data['scencmd'])
        self.op()

    @state_publisher(topic='SIMSETTINGS')
    def pub_simsettings(self):
        ''' Publish simulation settings on request. '''
        return dict(settings=bs.settings._settings_hierarchy,
            plugins=list(plugin.Plugin.plugins.keys()))

    def setutc(self, *args):
        ''' Set simulated clock time offset. '''
        if not args:
            pass  # avoid error message, just give time

        elif len(args) == 1:
            if args[0].upper() == 'RUN':
                self.utc = datetime.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

            elif args[0].upper() == 'REAL':
                self.utc = datetime.datetime.today().replace(microsecond=0)

            elif args[0].upper() == 'UTC':
                self.utc = datetime.datetime.utcnow().replace(microsecond=0)

            else:
                try:
                    self.utc = datetime.datetime.strptime(args[0], 
                        '%H:%M:%S.%f' if '.' in args[0] else '%H:%M:%S')
                except ValueError:
                    return False, 'Input time invalid'

        elif len(args) == 3:
            day, month, year = args
            try:
                self.utc = datetime.datetime(year, month, day)
            except ValueError:
                return False, 'Input date invalid.'
        elif len(args) == 4:
            day, month, year, timestring = args
            try:
                self.utc = datetime.datetime.strptime(
                    f'{year},{month},{day},{timestring}',
                    '%Y,%m,%d,%H:%M:%S.%f' if '.' in timestring else
                    '%Y,%m,%d,%H:%M:%S')
            except ValueError:
                return False, 'Input date invalid.'
        else:
            return False, 'Syntax error'

        return True, 'Simulation UTC ' + str(self.utc)

    @staticmethod
    def setseed(value):
        ''' Set random seed for this simulation. '''
        seed(value)
        np.random.seed(value)



from bluesky.network.common import get_ownip, seqidx2id, seqid2idx, GROUPID_CLIENT, GROUPID_DEFAULT, GROUPID_NOGROUP, GROUPID_SIM
from bluesky.network.subscriber import subscriber, subscribe
from bluesky.network.publisher import state_publisher, StatePublisher



''' Traffic-related classes. '''
from .traffic import Traffic
from .route import Route
from .activewpdata import ActiveWaypoint
from .adsbmodel import ADSB
from .autopilot import Autopilot
from .aporasas import APorASAS
from .turbulence import Turbulence
from .windfield import Windfield
from .windsim import WindSim


import sys
import bluesky as bs
from bluesky import cmdargs


def main():
    """
        Start BlueSky: This is the main entrypoint for BlueSky.
        Depending on settings and arguments passed it can start in different
        modes. The central part of BlueSky consists of a server managing all
        simulations, normally together with a gui. The different modes for this
        are:
        - server-gui: Start gui and simulation server
        - server-headless: start server without gui
        - client: start gui only, which can connect to an already running server

        A BlueSky server can start one or more simulation processes, which run
        the actual simulations. These simulations can also be started completely
        separate from all other BlueSky functionality, in the detached mode.
        This is useful when calling bluesky from within another python
        script/program. The corresponding modes are:
        - sim: The normal simulation process started by a BlueSky server
        - sim-detached: An isolated simulation node, without networking
    """
    # When importerror gives different name than (pip) install needs,
    # also advise latest version
    missingmodules = {"OpenGL": "pyopengl"}


    # Catch import errors
    try:
        # Parse command-line arguments
        args = cmdargs.parse()
        # Initialize bluesky modules. Pass command-line arguments parsed by cmdargs
        bs.init(**args)

        # Only start a simulation node if called with --sim or --detached
        if bs.mode == 'sim':
            bs.net.connect()
            bs.sim.run()
        else:
            # Only print start message in the non-sim cases to avoid printing
            # this for every started node
            print("   *****   BlueSky Open ATM simulator *****")
            print("Distributed under GNU General Public License v3")

        # Start server if server/gui or server-headless is started here
        if bs.mode == 'server':
            if bs.gui is None:
                bs.server.run()
            else:
                bs.server.start()

        # Start gui if client or main server/gui combination is started here
        if bs.gui == 'qtgl':
            from bluesky.ui import qtgl
            qtgl.start(hostname=args.get('hostname'))

        elif bs.gui == 'console':
            from bluesky.ui import console
            console.start(hostname=args.get('hostname'))

    # Give info on missing module
    except ImportError as error:
        modulename = missingmodules.get(error.name) or error.name
        if modulename is None or 'bluesky' in modulename:
            raise error
        print("Bluesky needs", modulename)
        print("Run setup-python.bat (Windows) or check requirements.txt (other systems)")
        print("Install using e.g. pip install", modulename)

    print('BlueSky normal end.')

if __name__ == '__main__':
    sys.exit(main())
